"""
This module contains Edge and Arrow subclasses for this behavioral package.
"""

from ....logger import logger
from ...graph import Edge
from ..execution.entities import Process
from ..filesystem.entities import File
from .entities import *

__all__ = ["HasImport", "IsFile"]





logger.info("Loading relations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class HasImport(Edge):

    """
    This kind of edge indicates that a process imported something.
    """

    label : str = ""

    source : Process
    destination : Import





class IsFile(Edge):

    """
    This kind of edge indicates that an import corresponds to an existing file object.
    """

    label : str = ""

    source : Import
    destination : File





del Edge, File, Import, Process, logger


"""
This is a *magic* script to bake baguettes. Use -h/--help for usage.
"""





def main():
    """
    Command line function to bake Cuckoo report(s). Use -h/--help for more info.
    """

    if __name__ in ("__main__", "bake"):        # One for when started with Python, the other for when started with package command...

        import logging
        from baguette.bakery.logger import logger, set_level
        set_level(logging.ERROR)

        import argparse
        from os import environ
        import pathlib
        from typing import Literal, Iterator
        from baguette.bakery.compiler import compile
        from baguette.bakery.source import filters
        from baguette.bakery.source.colors import Color
        from baguette.rack import BaguetteRack, TimeoutExit

        parser = argparse.ArgumentParser(
            description = 'Bakes Cuckoo reports into baguettes (gexf and pyt graphs).',
            add_help = True,
            conflict_handler = 'resolve',
            epilog = """
            Note that the format of input and outputs is quite flexible. Indeed you can :
            give output folders for each output file, which will make all input have their output in these folders
            - give a set of outputs for each input file, as long as they are given in the same place in the parameter sequences
            - do the same with input folders and output folders, as long as they also come in the same order
            - mix these options, again, as long as it has a meaning in the order they come.
            - Note that if a given path does not exists, it is interpreted as a folder.
            - Also, if an input path has the appropriate extension, it will be considered as a single input.
            """
            )

        class PathSorter:

            def __init__(self, pool : list[pathlib.Path | Literal["-"]], name : str) -> None:
                self.pool = pool
                self.name = name
            
            def __call__(self, arg : str):
                if arg == "-":
                    self.pool.append(arg)
                    return
                from glob import iglob
                import re
                magic_check = re.compile('([*?[])')
                try:
                    if magic_check.search(arg) is not None:
                        self.pool.extend(pathlib.Path(path) for path in iglob(arg))
                    else:
                        self.pool.append(pathlib.Path(arg))
                except:
                    parser.error("invalid {} path : '{}'".format(self.name, arg))

        reports : list[pathlib.Path | Literal["-"]] = []
        baguettes : list[pathlib.Path | Literal["-"]] = []
        visuals : list[pathlib.Path | Literal["-"]] = []
        outputs : list[pathlib.Path | Literal["-"]] = []

        def pool_size(arg : str) -> int:
            """
            Transforms a numeric argument in a number of process to use as a process pool.
            It can be absolute, negative (relative to the number of CPUs) or 
            """
            from os import cpu_count
            N = cpu_count()
            if not N:
                N = 1
            try:
                v = int(arg)
                if v < 0:
                    v = N - v
                if v <= 0:
                    parser.error("got a (too) negative value for process pool size : '{}'".format(arg))
            except:
                try:
                    v = float(arg)
                    if v <= 0:
                        parser.error("got a negative relative process pool size : '{}'".format(arg))
                    v = round(v * N)
                except:
                    parser.error("not a process pool size : '{}'".format(arg))
            return v

        def time(arg : str) -> float:
            from math import isnan
            try:
                v = float(arg)
                if v <= 0 or isnan(v):
                    parser.error("got a negative, null or nan maxtime")
                return v
            except:
                parser.error("expected positive float for maxtime, got : '{}'".format(arg))

        def paint_color(c : str) -> Color:
            if c in dir(Color):
                color = getattr(Color, c)
                if not isinstance(color, Color):
                    parser.error(f"not a valid color name : '{c}'")
                return color
            else:
                # We can match colors in ALLL LANGUAGES!!!!
                import re
                color_re = r"([\d\.eE-]+|\d+|0[xX][\daAbBcCdDeEfF]+)"
                sep_re = r"(?:[ ,;:\|\&]+)"
                inner_expr = r"(?:" + color_re + sep_re + color_re + sep_re + color_re + r")"
                fmatch = re.compile(inner_expr).fullmatch(c) or re.compile(r"\(" + inner_expr + r"\)").fullmatch(c) or re.compile(r"\[" + inner_expr + r"\]").fullmatch(c) or re.compile(r"\{" + inner_expr + r"\}").fullmatch(c)
                if not fmatch:
                    parser.error(f"could not understand color format : '{c}'")
                r, g, b = fmatch.groups()

                def is_float(s : str) -> bool:
                    try:
                        f = float(s)
                        if not 0 <= f <= 1:
                            return False
                        return True
                    except:
                        return False
                
                def is_int(s : str) -> bool:
                    try:
                        i = int(s)
                        if not 0 <= i <= 255:
                            return False
                        return True
                    except:
                        return False
                
                def is_hex(s : str) -> bool:
                    try:
                        i = int(s, base=16)
                        if not 0 <= i <= 255:
                            return False
                        return True
                    except:
                        return False
                    
                if all(is_float(x) for x in (r, g, b)):
                    r, g, b = float(r), float(g), float(b)
                elif all(is_int(x) for x in (r, g, b)):
                    r, g, b = int(r), int(g), int(b)
                elif all(is_hex(x) for x in (r, g, b)):
                    r, g, b = int(r, 16), int(g, 16), int(b, 16)
                else:
                    parser.error(f"could not understand color format : '{c}'")

                return Color(r, g, b)

        parser.add_argument("reports", type=PathSorter(reports, "report"), default=None, nargs="*" if "BAGUETTE_REPORTS" in environ else "+", help="Cuckoo report files (.json) to bake baguettes from. Can also be folders containing reports. Defaults to environment variable 'BAGUETTE_REPORTS' if set.")
        parser.add_argument("--baguettes", type=PathSorter(baguettes, "baguette"), default=None, action="extend", nargs="*", help="the path(s) to the output baguette files (.pyt). Use '-' to leave it to the automatic destination. Defaults to environment variable 'BAGUETTE_BAGUETTES' or '-' if not set.")
        parser.add_argument("--visuals", type=PathSorter(visuals, "visual"), default=None, action="extend", nargs="*", help="the path(s) to the output visual (Gephi) files (.gexf). Use '-' to leave it to the automatic destination. Defaults to environment variable 'BAGUETTE_VISUALS' or '-' if not set.")
        parser.add_argument("-o", "--outputs", type=PathSorter(outputs, "output"), default=None, action="extend", nargs="*", help="the path to the result index folders (which end in .bag). They contain the index file (.pyt) which stores all the information about a given baguette. Use '-' to leave it to the automatic destination. Defaults to environment variable 'BAGUETTE_OUTPUTS' or '.' if not set.")
        parser.add_argument("--pool", type=pool_size, default=pool_size("0.5"), help="the size of the process pool to use to bake in parallel.")
        parser.add_argument("--maxtime", type=time, default=time("inf"), help="the maximum amount of time spent baking a single baguette. No maxtime by default.")
        parser.add_argument("-f", "--filters", type=str, default=[], choices=[name for name in dir(filters) if isinstance(getattr(filters, name), filters.Filter)], nargs="*", help="a list of filters that can be used when exporting the baguette to the visual file (.gexf).")
        parser.add_argument("--perf", action="store_true", default=False, help="if this is enabled, a performance report will be printed at the end of the baking process.")
        parser.add_argument("-v", "--verbosity", action="count", default=0, help="increases the verbosity of the output.")
        parser.add_argument("--skip_data_comparison", action="store_true", default=False, help="if enabled, the computation of the Levenshtein similarity between all Data nodes will be skipped.")
        parser.add_argument("--skip_diff_comparison", action="store_true", default=False, help="same as skip_data_comparison but for Diff nodes.")
        parser.add_argument("--background", type=paint_color, default=Color.black, help="If a color is given for background, the color settings which are close to the background color will be change to be more visible on background. Must be a valid color name or RGB values.")


        args = parser.parse_args()

        # Setting logging level

        levels = {
            0 : logging.ERROR,
            1 : logging.WARNING,
            2 : logging.INFO,
            3 : logging.DEBUG
        }
        verbosity : Literal[0, 1, 2, 3] = min(3, args.verbosity)
        set_level(levels[verbosity])

        logger.info("Arguments parsed. Discovering jobs.")

        # Parsing jobs

        if not reports:
            try:
                reports = [pathlib.Path(environ["BAGUETTE_REPORTS"])]
            except KeyError:
                raise RuntimeError("Environment variable 'BAGUETTE_REPORTS' not found and there is no input...")
            except:
                parser.error("invalid report path in environment variable : '{}'".format(environ["BAGUETTE_REPORTS"]))
        if not baguettes:
            if "BAGUETTE_BAGUETTES" in environ:
                try:
                    baguettes = [pathlib.Path(environ["BAGUETTE_BAGUETTES"])]
                except:
                    parser.error("invalid baguettes path in environment variable : '{}'".format(environ["BAGUETTE_BAGUETTES"]))
            else:
                baguettes = ["-"]
        if not visuals:
            if "BAGUETTE_VISUALS" in environ:
                try:
                    visuals = [pathlib.Path(environ["BAGUETTE_VISUALS"])]
                except:
                    parser.error("invalid visuals path in environment variable : '{}'".format(environ["BAGUETTE_VISUALS"]))
            else:
                visuals = ["-"]
        if not outputs:
            if "BAGUETTE_OUTPUTS" in environ:
                try:
                    outputs = [pathlib.Path(environ["BAGUETTE_OUTPUTS"])]
                except:
                    parser.error("invalid outputs path in environment variable : '{}'".format(environ["BAGUETTE_OUTPUTS"]))
            else:
                outputs = [pathlib.Path(".")]

        class JobQueue:

            """
            These objects hold a series of grouped jobs.
            """

            def __init__(self, groups : list[pathlib.Path | Literal["-"] | list[pathlib.Path]] = []) -> None:
                from copy import deepcopy
                self.__groups : list[pathlib.Path | Literal["-"] | list[pathlib.Path]] = deepcopy(groups)
                self.__last_group : list[pathlib.Path] = []
                self.__active : bool = False
            
            def insert_group(self, group : list[pathlib.Path] | pathlib.Path | Literal["-"]):
                """
                Creates an independant work group with given content, also creating a new active group for later use.
                """
                if self.__active:
                    self.__groups.append(self.__last_group)
                self.__groups.append(group)
                self.__last_group = []
                self.__active = False
            
            def append_to_active_group(self, path : pathlib.Path):
                """
                Appends a file to the current group.
                """
                self.__last_group.append(path)
                self.__active = True
            
            def new_group(self):
                """
                Adds the active group to the queue and creates a new empty active group.
                """
                if self.__active:
                    self.__groups.append(self.__last_group)
                self.__last_group = []
                self.__active = True

            def finalize(self):
                """
                Finalizes the queue, adding the currently active group to the queue if necessary.
                """
                if self.__active and self.__last_group:
                    self.__groups.append(self.__last_group)
            
            def __iter__(self) -> Iterator[pathlib.Path | Literal["-"] | list[pathlib.Path]]:
                """
                Yields all the jobs in the queue.
                """
                yield from self.__groups
            
            def __len__(self) -> int:
                """
                Implements len(self).
                """
                return len(self.__groups)

            def __getitem__(self, index : int) -> pathlib.Path | Literal["-"] | list[pathlib.Path]:
                """
                Implements self[index].
                """
                return self.__groups[index]


        report_groups : JobQueue = JobQueue()
        baguette_groups : JobQueue = JobQueue()
        visual_groups : JobQueue = JobQueue()
        output_groups : JobQueue = JobQueue()

        def determine_type(path : pathlib.Path | Literal["-"], ext : str) -> Literal["folder", "file"]:
            """
            Determines if the given path should be interpreted as a folder or file path.
            If it ends with the given extension and does not exist, it well be considered a file.
            """
            if path == "-":
                return "folder"
            elif ext and path.suffix == ext:
                return "file"
            if path.is_file():
                return "file"
            elif path.is_dir():
                return "folder"
            elif path.exists():
                parser.error("given path exists and is neither a file or folder : '{}'".format(path))
            else:
                return "folder"
        
        def create_name(folder : pathlib.Path) -> pathlib.Path:
            """
            Given a path, this will return a (possibly) modified path that does not exist in the same folder.
            """
            if not folder.exists():
                return folder
            n = 0
            ext = folder.suffix
            if ext:
                sfolder = str(folder)[:-len(ext)]
            else:
                sfolder = str(folder)
            new_folder = pathlib.Path(sfolder + "({})".format(n) + ext)
            while new_folder.exists():
                n += 1
                new_folder = pathlib.Path(sfolder + "({})".format(n) + ext)
            return new_folder
            

        for r in reports:
            if r == "-":
                parser.error("cannot use automatic destination operator for input files.")
            if determine_type(r, ".json") == "folder":
                if r.exists():
                    report_groups.insert_group([p for p in r.iterdir()])
                else:
                    parser.error("input folder/file does not exist : '{}'".format(r))
            else:
                if r.exists():
                    report_groups.append_to_active_group(r)
                else:
                    parser.error("input file does not exist : '{}'".format(r))
        report_groups.finalize()
        
        for b in baguettes:
            if determine_type(b, ".pyt") == "folder":
                baguette_groups.insert_group(b)
            else:
                if b == "-":
                    raise RuntimeError("How?")
                baguette_groups.append_to_active_group(b)
        baguette_groups.finalize()
        if len(baguette_groups) == 1 and len(report_groups) > 1 and (isinstance(baguette_groups[0], pathlib.Path) or baguette_groups[0] == "-"):
            baguette_groups = JobQueue([baguette_groups[0]] * len(report_groups))

        for v in visuals:
            if determine_type(v, ".gexf") == "folder":
                visual_groups.insert_group(v)
            else:
                if v == "-":
                    raise RuntimeError("How?")
                visual_groups.append_to_active_group(v)
        visual_groups.finalize()
        if len(visual_groups) == 1 and len(report_groups) > 1 and (isinstance(visual_groups[0], pathlib.Path) or visual_groups[0] == "-"):
            visual_groups = JobQueue([visual_groups[0]] * len(report_groups))
        
        for o in outputs:
            if determine_type(o, ".bag") == "folder":
                output_groups.insert_group(o)
            else:
                if o == "-":
                    raise RuntimeError("How?")
                output_groups.append_to_active_group(o)
        output_groups.finalize()
        if len(output_groups) == 1 and len(report_groups) > 1 and (isinstance(output_groups[0], pathlib.Path) or output_groups[0] == "-"):
            output_groups = JobQueue([output_groups[0]] * len(report_groups))

        # def printable_format(l) -> str:
        #     return str(l) if not isinstance(l, list) else (str(l) if len(l) < 2 else "[...]")

        # print("Sizes: {}, {}, {}, {}".format(len(report_groups), len(baguette_groups), len(visual_groups), len(output_groups)))
        # print(list(printable_format(r) for r in report_groups))
        # print(list(printable_format(b) for b in baguette_groups))
        # print(list(printable_format(v) for v in visual_groups))
        # print(list(printable_format(o) for o in output_groups))
        # l1 = list(report_groups)
        # l2 = list(baguette_groups) + [None] * (len(report_groups) - len(baguette_groups))
        # l3 = list(visual_groups) + [None] * (len(report_groups) - len(visual_groups))
        # l4 = list(output_groups) + [None] * (len(report_groups) - len(output_groups))
        # for r, b, v, o in zip(l1, l2, l3, l4):
        #     print("Work group:")
        #     print("reports :", printable_format(r))
        #     print("baguettes :", printable_format(b))
        #     print("visuals :", printable_format(v))
        #     print("outputs :", printable_format(o))
        #     print("\n")

        if len(report_groups) != len(baguette_groups) or len(report_groups) != len(visual_groups) or len(report_groups) != len(output_groups):
            parser.error("different number of work groups in inputs/outputs.")
        
        for r, b, v, o in zip(report_groups, baguette_groups, visual_groups, output_groups):
            if not isinstance(r, list):
                raise RuntimeError("How did we get here?")
            n = len(r)
            if isinstance(b, list) and len(b) != n:
                parser.error("got a work group with different numbers of inputs and outputs.")
            if isinstance(v, list) and len(v) != n:
                parser.error("got a work group with different numbers of inputs and outputs.")
            if isinstance(o, list) and len(o) != n:
                parser.error("got a work group with different numbers of inputs and outputs.")

        work : list[BaguetteRack] = []

        def stripname(p : pathlib.Path) -> str:
            """
            Returns the name of the path without the exetension if it has one.
            """
            ext = p.suffix
            if ext:
                return p.name[:-len(ext)]
            return p.name

        for rgroup, bgroup, vgroup, ogroup in zip(report_groups, baguette_groups, visual_groups, output_groups):

            if rgroup == "-":
                parser.error("cannot use automatic destination operator for input files.")
            elif isinstance(rgroup, pathlib.Path):      # Folder input
                rgroup = list(rgroup.iterdir())         # Transform in file inputs

            if ogroup == "-":
                parser.error("cannot use automatic destination operator for output folders.")
            elif isinstance(ogroup, pathlib.Path):      # Folder output
                ogroup = [create_name(pathlib.Path(ogroup, stripname(r) + ".bag")) for r in rgroup]
            
            if bgroup == "-":                           # Magic destination:
                bgroup = [None for _ in rgroup]         # To be determined later
            elif isinstance(bgroup, pathlib.Path):      # Folder baguette destination
                bgroup = [create_name(pathlib.Path(bgroup, stripname(r) + ".pyt")) for r in rgroup]       # Pre-compute the baguette paths with report names

            if vgroup == "-":                           # Magic destination:
                vgroup = [None for _ in rgroup]         # To be determined later
            elif isinstance(vgroup, pathlib.Path):      # Folder visual destination
                vgroup = [create_name(pathlib.Path(vgroup, stripname(r) + ".gexf")) for r in rgroup]       # Pre-compute the visual paths with report names
            
            for r, b, v, o in zip(rgroup, bgroup, vgroup, ogroup):
                bg = BaguetteRack(o)
                bg.report = r
                bg.baguette = b if b is not None else (bg.working_directory / "baguette.pyt")
                bg.visual = v if v is not None else (bg.working_directory / "visual.gexf")
                bg.verbosity = verbosity
                bg.skip_data_comparison = args.skip_data_comparison
                bg.skip_diff_comparison = args.skip_diff_comparison
                bg.filter_names = args.filters
                bg.maxtime = args.maxtime
                bg.perf = args.perf
                bg.background_color = args.background
                work.append(bg)
        
        
        # Compile now...

        from multiprocessing.pool import Pool
        from threading import Lock, Thread

        # All of this is because multiprocessing was coded with feet... Pool's async methods may freeze (deadlock maybe) on some platforms.

        lock = Lock()
        failed, timed_out, total = 0, 0, len(work)
        def execute_single_job(P : Pool) -> bool:
            nonlocal failed, timed_out
            with lock:
                if not work: 
                    return False
                br = work.pop()
            try:
                br = P.apply(compile, (br, ))
            except KeyboardInterrupt as e:
                from traceback import TracebackException
                br.exception = TracebackException.from_exception(e)
            if br.exception is None or not br.suppressed:
                br.export()
            if br.exception is not None and issubclass(br.exception.exc_type, KeyboardInterrupt):
                return False
            elif br.exception is not None and not issubclass(br.exception.exc_type, TimeoutExit):
                with lock:
                    failed += 1
                logger.error("Got a '{}' error during the baking of '{}'.".format(br.exception.exc_type.__name__, br.report.name))
            elif br.exception is not None and issubclass(br.exception.exc_type, TimeoutExit):
                with lock:
                    timed_out += 1
            return True

        def executor(P : Pool):
            while execute_single_job(P):
                pass
        
        threads : list[Thread] = []
        try:
            with Pool(args.pool, maxtasksperchild = 1) as P:
                for _ in range(args.pool):
                    t = Thread(target = executor, args = (P, ), daemon = True)
                    t.start()
                    threads.append(t)

                for t in threads:
                    t.join()
            
            success = total - failed - timed_out
            if failed and success and timed_out:
                print("{} failed baguettes, {} took too long and {} well-baked.".format(failed, timed_out, success))
            elif failed and success:
                print("{} failed baguettes, {} baked correctly.".format(failed, success))
            elif timed_out and success:
                print("{} baguettes took too long to bake, {} baked correctly.".format(timed_out, success))
            elif failed and timed_out:
                print("{} baguettes are failed and the {} others took too long to bake...".format(failed, timed_out))
            elif failed:
                print("All {} baguettes did not bake correctly...".format(failed))
            elif timed_out:
                print("All {} baguettes took too long to bake...".format(timed_out))
            elif success:
                print("All {} are well-baked!".format(success))
        except KeyboardInterrupt:
            print("Exiting.")
            




if __name__ == "__main__":
    main()


"""
This module contains Vertex subclasses for this behavioral package.
"""

from typing import Iterator, Optional

from ....logger import logger
from ...config import ColorSetting, SizeSetting
from ...colors import Color
from ...graph import UniqueVertex, Vertex
from ...utils import chrono
from ..filesystem.entities import File, Handle
from ..network.entities import Connection, Socket
from .utils import IOOperation

__all__ = ["Data", "Diff"]





logger.info("Loading entities from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class Data(UniqueVertex):
    
    """
    A data vertex. Represents data read, written or appended to a file.
    """

    similarity_threshold = 0.75

    __slots__ = {
        "data" : "The bytes data exchanged through the handle",
        "length" : "The length of the data in bytes",
        "entropy" : "The byte-wise entropy of the message",
        "time" : "The time at which this message was seen",
        "isprintable" : "Indicates if a message can be converted to a UTF-8 str and contains only printable characters",
        "__initialized" : "Indicates if this Vertex can be compared to other Data Vertices"
    }

    __pickle_slots__ = {
        "data",
        "length",
        "entropy",
        "time",
        "isprintable"
    }

    default_color = ColorSetting(Color(0.5882352941176471, 1.0, 1.0))
    default_size = SizeSetting(0.5)

    def __init__(self, *, parent: Optional[Vertex] = None) -> None:
        super().__init__(parent=parent)
        self.__initialized = False
        self.data : bytes = b""
        self.length : int = 0
        self.entropy : float = 0.0
        self.time : float = 0.0
        self.isprintable : bool = False
    
    @property
    def vector(self) -> Handle | Socket:
        """
        The Handle or Socket Vertex that this Data node is Conveying data to or from.
        """
        from ..filesystem import Handle
        from ..network import Socket
        for u in self.neighbors():
            if isinstance(u, Handle | Socket):
                return u
        raise RuntimeError("Got a data node without vector.")
    
    @chrono
    def set_data(self, data : str):
        """
        Writes the corresponding data to the node.
        Also computes all related metrics.
        """
        from .utils import entropy
        self.data = data.encode("utf-8")
        self.length = len(self.data)
        self.entropy = entropy(self.data)
        self.isprintable = data.isprintable()
    
    def compare(self):
        """
        Links the node to other similar nodes.
        """
        from .relations import IsSimilarTo
        from .utils import levenshtein_similarity
        self.__initialized = True
        for d in Data:
            if (1 - Data.similarity_threshold) * max(len(d.data), len(self.data)) < abs(len(d.data) - len(self.data)):      # Size difference is too high, similarity will be below threshold
                continue
            if d is not self and d.__initialized:
                s1 = levenshtein_similarity(d.data, self.data, Data.similarity_threshold)
                if s1 < Data.similarity_threshold:
                    continue
                if self.time < d.time:
                    l1 = IsSimilarTo(self, d)
                else:
                    l1 = IsSimilarTo(d, self)
                l1.weight = s1
                # s2 = levenshtein_subset_similarity(d.data, self.data)
                # if self.time < d.time:
                #     l2 = IsAlmostIn(self, d)
                # else:
                #     l2 = IsAlmostIn(d, self)
                # l2.weight = s2





class Diff(UniqueVertex):

    """
    A diff vertex. This represents all the information gathered on a file's content during the lifetime of a handle.
    """

    similarity_threshold = 0.90

    __slots__ = {
        "read" : "The content of the read diff file",
        "read_type" : "The file type determined by libmagic from what has been read",
        "written" : "The content of the write diff file",
        "written_type" : "The file type determined by libmagic from what has been written",
        "glob" : "The content of the global diff file",
        "glob_type" : "The file type determined by libmagic from the final state of the file",
        "read_total" : "The total amount of bytes that were read",  # Counts double if you read twice the same byte
        "read_space" : "The amount of bytes that were read in the file",    # This one only counts one
        "written_total" : "The total amount of bytes that were written",
        "written_space" : "The amount of bytes that were written in the file",
        "glob_space" : "The amount of bytes that were accessed in the file",
        "read_entropy" : "The amount of entropy that was read from the file",
        "written_entropy" : "The amount of entropy that was written to the file",
        "glob_entropy" : "The entropy that resulted in the file from all operations",
        "printable_rate" : "Indicates how much of the final state of the file only contains printable characters",
        "encoding" : "A valid encoding for the final file",
        "__initialized" : "Indicates if this Vertex can be compared to other Diff Vertices",
        "__reader_difffile" : "The DiffFile that will store the result of all reading operations",
        "__writer_difffile" : "The DiffFile that will store the result of all writing operations",
        "__glob_difffile" : "The DiffFile that will store the result of all IO operations",
        "__operations" : "The list of all operations that appear in the Diff node"
    }

    __pickle_slots__ = {
        "read",
        "read_type",
        "written",
        "written_type",
        "glob",
        "glob_type",
        "read_total",
        "read_space",
        "written_total",
        "written_space",
        "glob_space",
        "read_entropy",
        "written_entropy",
        "glob_entropy",
        "printable_rate",
        "encoding",
    }

    _active_target_diff : dict[File | Connection, "Diff"] = {}
    _active_vector_diff : dict[Handle | Socket, "Diff"] = {}

    default_color = ColorSetting(Color(50, 150, 255))
    default_size = SizeSetting(1.5)

    min_size = SizeSetting(0.5)
    max_size = SizeSetting(2.5)

    diff_low_entropy_color = ColorSetting(Color(50, 150, 255))
    diff_high_entropy_color = ColorSetting(Color(255, 50, 150))

    def __init__(self, *, parent: Optional["Vertex"] = None) -> None:
        from .utils import DiffFile, IOOperation
        super().__init__(parent=parent)
        self.__initialized : bool = False
        self.__reader_difffile = DiffFile()
        self.__writer_difffile = DiffFile()
        self.__glob_difffile = DiffFile()
        self.__operations : list[IOOperation] = []
        self.read : bytes = b""
        self.read_type : list[str] = []
        self.written : bytes = b""
        self.written_type : list[str] = []
        self.glob : bytes = b""
        self.glob_type : list[str] = []
        self.read_total : int = 0
        self.read_space : int = 0
        self.written_total : int = 0
        self.written_space : int = 0
        self.glob_space : int = 0
        self.read_entropy : float = 0.0
        self.written_entropy : float = 0.0
        self.glob_entropy : float = 0.0
        self.printable_rate : float = 0.0
        self.encoding : str = "raw"

    @property
    def read_buffer(self) -> bytes:
        """
        The current content of the read diff file.
        """
        return self.__reader_difffile.dump()
    
    @property
    def write_buffer(self) -> bytes:
        """
        The current content of the write diff file.
        """
        return self.__writer_difffile.dump()
    
    @property
    def glob_buffer(self) -> bytes:
        """
        The current content of the global diff file.
        """
        return self.__glob_difffile.dump()
    
    @property
    def content(self) -> tuple[bytes, bytes, bytes]:
        """
        The current content of the read, write and global diff files.
        """
        return self.read_buffer, self.write_buffer, self.glob_buffer
        
    def last_pos(self) -> int:
        """
        Returns the offset after the las operation performed in this Diff node.
        """
        if self.__operations:
            return self.__operations[-1].stop
        else:
            return 0
    
    @chrono
    def add_operation(self, op : IOOperation):
        """
        Registers a new IO operation to this Diff node.
        """
        from .utils import Read, Write
        self.__operations.append(op)
        self.__glob_difffile.write(op.data, op.start)
        if isinstance(op, Read):
            self.__reader_difffile.write(op.data, op.start)
        elif isinstance(op, Write):
            self.__writer_difffile.write(op.data, op.start)
    
    @chrono
    def compute_data(self):
        """
        Computes the values of all the attributes given the state of operations and DiffFiles.
        """
        from magic import Magic

        from ...config import CompilationParameters
        from ...graph import Graph
        from .relations import (HasSimilarContent, IsDiffOf, IsReadBy,
                                WritesInto)
        from .utils import (Read, Write, entropy, levenshtein_similarity,
                            printable_rate)

        analyzer = Magic(keep_going=True, uncompress=True)
        r, w, g = self.content
        self.read, self.written, self.glob = r, w, g
        self.read_type = analyzer.from_buffer(r).split("\\012- ")
        self.written_type = analyzer.from_buffer(w).split("\\012- ")
        self.glob_type = analyzer.from_buffer(g).split("\\012- ")
        self.read_total = sum(len(op.data) for op in self.__operations if isinstance(op, Read))
        self.written_total = sum(len(op.data) for op in self.__operations if isinstance(op, Write))
        self.read_space = len(r)
        self.written_space = len(w)
        self.glob_space = len(g)
        self.read_entropy = entropy(r)
        self.written_entropy = entropy(w)
        self.glob_entropy = entropy(g)
        self.printable_rate, self.encoding = printable_rate(g)

        if not r and w:
            for e in self.edges:
                if isinstance(e, IsDiffOf):
                    WritesInto(e.source, e.destination)
                    e.delete()
                    for G in Graph.active_graphs():
                        G.remove(e)
        elif not w and r:
            for e in self.edges:
                if isinstance(e, IsDiffOf):
                    IsReadBy(e.destination, e.source)
                    e.delete()
                    for G in Graph.active_graphs():
                        G.remove(e)

        if CompilationParameters.SkipLevenshteinForDiffNodes:
            return
        
        for u in Diff:
            if u not in self.neighbors() and u is not self:
                lw, ls, ld = 0.0, "", ""
                for sn, sb in zip(("read_buffer", "write_buffer", "global_buffer"), (r, w, g)):
                    if sb:
                        for un, ub in zip(("read_buffer", "write_buffer", "global_buffer"), u.content):
                            if (1 - Diff.similarity_threshold) * max(len(sb), len(ub)) < abs(len(sb) - len(ub)):      # Size difference is too high, similarity will be below threshold
                                continue
                            if ub:
                                s = levenshtein_similarity(sb, ub, Diff.similarity_threshold)
                                if s >= lw:
                                    lw = s
                                    ls = sn
                                    ld= un
                if ls and lw >= Diff.similarity_threshold:      # Heuristic is not perfect : checking that the threshold has indeed been reached!
                    l = HasSimilarContent(self, u)
                    l.weight = lw * (l.max_weight - l.min_weight) + l.min_weight        # 0 <= lw <= 1
                    l.source_buffer = ls
                    l.destination_buffer = ld
    
    @property
    def label(self) -> str:
        """
        The name of the node to display. It is the global data type.
        """
        if len(self.glob_type) == 1:
            return self.glob_type[0]
        return " | ".join(t for t in self.glob_type if t != "data")

    @property
    def vectors(self) -> Iterator[Handle | Socket]:
        """
        The Handle or Socket Vertices that this Diff node is interacting with.
        """
        from ..filesystem import Handle
        from ..network import Socket
        for u in self.neighbors():
            if isinstance(u, Handle | Socket):
                yield u
    
    @property
    def targets(self) -> Iterator[File | Connection]:
        """
        The File or Connection Vertices that this Diff node is interacting with.
        """
        from ..filesystem import File
        from ..network import Connection
        for u in self.neighbors():
            if isinstance(u, File | Connection):
                yield u





del Color, ColorSetting, SizeSetting, Connection, File, Handle, IOOperation, Iterator, Optional, Socket, UniqueVertex, Vertex, chrono, logger


"""
This module contains Vertex subclasses for this behavioral package.
"""

from typing import Optional

from ....logger import logger
from ...colors import Color
from ...config import ColorSetting, SizeSetting
from ...graph import UniqueVertex, Vertex
from ...utils import chrono

__all__ = ["Process", "Thread", "Call"]





logger.info("Loading entities from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

CommandTree = list[tuple[str, "CommandTree"]]

class Process(UniqueVertex):

    """
    A process vertex. Holds information to identify a process.
    """

    __slots__ = {
        "PID" : "The PID of the process during its execution",
        "__command" : "The command that this process is running",
        "executable" : "The file system path that the process was started in",
        "start" : "The time at which the process was started",
        "stop" : "The time at chich the process was stopped",
        "__sub_commands" : "The tree of commands executed by all chile processes"
    }

    __pickle_slots__ = {
        "PID",
        "command",
        "executable",
        "start",
        "stop",
        "sub_commands"
    }

    default_color = ColorSetting(Color(255, 255, 50))
    default_size = SizeSetting(5.0)

    def __init__(self, *, parent: Vertex | None = None) -> None:
        super().__init__(parent=parent)
        self.PID : int = 0
        self.__command : tuple[str] = tuple()
        self.executable : str = ""
        self.start : float = 0.0
        self.stop : float = 0.0
        self.__sub_commands : CommandTree | None = None

        from argparse import ArgumentParser
        from pathlib import PurePath

        from ...utils import is_path, path_factory
        from ..filesystem.integration import NewFile
        from .relations import Runs, UsesAsArgument

        def to_absolute(p : PurePath, cwd : PurePath) -> PurePath:
            if p.is_absolute():
                return p
            return cwd / p
        
        p = ArgumentParser()
        
        def link(e : NewFile):
            from ..filesystem import File
            f = e.file
            if not f.name:
                return
            # print("New file :", repr(f.path))
            # # print("Process :", repr(self.command))
            # print("Process #{} : '{}'".format(self.PID, self.executable))
            # print("Splitting command : {}".format(self.__command))
            for i, arg in enumerate(self.__command):
                if not i:
                    arg = self.executable
                while arg.endswith(" "):
                    arg = arg[:-1]
                while arg.startswith(" "):
                    arg = arg[1:]
                if arg.startswith("\"") and arg.endswith("\""):
                    arg = arg[1:-1]
                if arg.startswith("'") and arg.endswith("'"):
                    arg = arg[1:-1]
                # print(">>>", f.name.lower(), arg.lower())
                if (is_path(arg) and f.name.lower() == to_absolute(path_factory(arg), path_factory(self.executable)).name.lower()) or f.name.lower() in arg.lower():
                # if (f.name in arg and len(f.name) / len(arg) > 0.9) or (str(f.path) in arg and len(str(f.path)) / len(arg) > 0.9):        # You need to work with a process cwd
                    if i > 0:
                        UsesAsArgument(self, f)
                        break
                    else:
                        Runs(self, f)

        NewFile.add_callback(link)

    @property
    def label(self) -> str:
        """
        Returns a label for the Process node.
        """
        return "Process #" + str(self.PID)
        
    @property
    def threads(self) -> list["Thread"]:
        """
        List of all the threads that this process had.
        """
        from .relations import HasThread
        return [e.destination for e in self.edges if isinstance(e, HasThread)] 
    
    @property
    def children_processes(self) -> list["Process"]:
        """
        List of all the children processes that this process created.
        """
        from .relations import HasChildProcess
        return [e.destination for e in self.edges if isinstance(e, HasChildProcess) and e.source is self]
    
    @property
    def parent_process(self) -> Optional["Process"]:
        """
        Returns the parent process node if one exists in the graph.
        """
        from .relations import HasChildProcess
        for e in self.edges:
            if isinstance(e, HasChildProcess) and e.destination is self:
                return e.source
            
    @property
    def command(self) -> str:
        """
        The command ran by this Process.
        """
        return " ".join(self.__command)
    
    @command.setter
    def command(self, cmd : str):
        from ...utils import parse_command_line
        if not isinstance(cmd, str):
            raise TypeError("Expected str, got " + repr(type(cmd).__name__))
        self.__command = tuple(parse_command_line(cmd))
        if not self.__command and self.executable:
            self.__command = (self.executable, )
        
    def parsed_command(self) -> list[str]:
        """
        Returns the argument vector used for starting this process.
        """
        return list(self.__command)

    @property
    def sub_commands(self) -> CommandTree:
        """
        Returns a dict structure that represents the commands executed by all child processes.
        """
        from .relations import HasChildProcess
        if self.__sub_commands != None:
            return self.__sub_commands
        sc = []
        for e in self.edges:
            if isinstance(e, HasChildProcess) and e.source is self:
                sc.append((e.destination.command, e.destination.sub_commands))
        self.__sub_commands = sc
        return sc
    
    @sub_commands.setter
    def sub_commands(self, value : CommandTree):
        self.__sub_commands = value





class Thread(UniqueVertex):

    """
    A thread vertex. Holds information to identify a thread.
    """

    __slots__ = {
        "TID" : "The TID of the thread during its execution",
        "Ncalls" : "The number of system calls that the thread made",
        "first" : "The first system call that this thread made",
        "last" : "the last system call that this thread made",
        "start" : "The time at which the thread was started",
        "stop" : "The time at chich the thread was stopped"
    }

    __pickle_slots__ = {
        "TID",
        "Ncalls",
        "first",
        "last",
        "start",
        "stop"
    }

    default_color = ColorSetting(Color(255, 204, 0))
    default_size = SizeSetting(2.0)

    def __init__(self, *, parent: Optional[Vertex] = None) -> None:
        super().__init__(parent=parent)
        self.TID : int = 0
        self.Ncalls : int = 0
        self.first : Call | None = None
        self.last : Call | None = None
        self.start : float = 0.0
        self.stop : float = 0.0

    @property
    def label(self) -> str:
        """
        Returns a label for the Thread node.
        """
        return "Thread #" + str(self.TID)
    
    @property
    def process(self) -> Process:
        """
        The Process Vertex that runs this Thread.
        """
        from .relations import HasThread
        for e in self.edges:
            if isinstance(e, HasThread):
                return e.source
        raise RuntimeError("Got a Thread without a parent Process.")





class Call(UniqueVertex):

    """
    A system call vertex. Holds information about a specific system call.
    Don't forget to call c.integrate() after setup of Call c is finished to integrate the call in the graph.
    """

    __slots__ = {
        "name" : "The name of the system call",
        "category" : "The family of the system call",
        "stacktrace" : "The stacktrace of the call",
        "status" : "Indicates if the call ran succesfully",
        "arguments" : "The arguments that the call received",
        "return_value" : "The value that the call returned",
        "time" : "The timestamp at which the call was detected",
        "flags" : "The flags of the call",
        "__thread" : "A shortcut to the Thread vertex that made this call"
    }

    __pickle_slots__ = {
        "name",
        "category",
        "stacktrace",
        "status",
        "arguments",
        "return_value",
        "time",
        "flags",
        "thread"
    }

    default_color = ColorSetting(Color(255, 153, 0))
    default_size = SizeSetting(0.3)

    def __init__(self, *, parent: Optional[Vertex] = None) -> None:
        from typing import Any

        from ...record import record
        super().__init__(parent=parent)
        self.name : str = ""
        self.category : str = ""
        self.stacktrace : tuple = ()
        self.status : int = 0
        self.arguments : record = record()
        self.return_value : Any = None
        self.time : float = 0.0
        self.flags : record = record()
        self.__thread : Thread | None = None
    
    @property
    def thread(self) -> Thread:
        """
        The Thread that made this Call.
        """
        if self.__thread is None:
            raise RuntimeError("Got a Call without a calling Thread.")
        return self.__thread
    
    @thread.setter
    def thread(self, value : Thread):
        if not isinstance(value, Thread):
            raise TypeError("Expected Thread, got " + repr(type(value).__name__))
        self.__thread = value

    @property
    def label(self) -> str:
        """
        The label for this node.
        """
        return self.name
    




del Color,ColorSetting, SizeSetting, CommandTree, Optional, UniqueVertex, Vertex, chrono, logger


"""
This package defines the graph-related subclasses that modelize importation mechanisms.
"""

from ....logger import logger

__all__ = ["entities", "relations", "integration"]





logger.info("Finalizing {} library.".format(__name__.rpartition(".")[2]))

from .entities import *
from .relations import *


from ....logger import logger





__proc__.init = ""
__proc__.term = "__init__"          # This is to trick most IDEs into thinking that the init code will be executed first, but it is only done at the end to avoid circular imports

__all__ = ["entities", "relations", "integration"]





logger.info("Initializing {} library.".format(__name__.rpartition(".")[2]))


"""
This module contains Edge and Arrow subclasses for this behavioral package.
"""

from ....logger import logger
from ...graph import Arrow, Edge
from ..execution.entities import Process
from ..filesystem.entities import Directory, File
from ..network.entities import Host
from .entities import Handle, Key, KeyEntry

__all__ = ["HasSubKey", "HasEntry", "UsesKey", "HasHandle", "Discovered", "QueriesEntry", "SetsEntry", "DeletesEntry", "ChangesTowards", "ReferencesFileSystem"]





logger.info("Loading relations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class HasSubKey(Arrow):

    """
    This kind of arrow indicates that a key has a sub-key.
    """

    label : str = ""

    source : Host | Key
    destination : Key





class HasEntry(Edge):

    """
    This kind of edge indicates that a registry key has one registry entry.
    """

    label : str = ""

    source : Key
    destination : KeyEntry





class UsesKey(Edge):

    """
    This kind of edge indicates that a handle uses a registry key.
    """

    label : str = ""

    source : Handle
    destination : Key





class HasHandle(Edge):

    """
    This kind of edge indicates that a process owns a registry key handle.
    """

    label : str = ""

    source : Process
    destination : Handle





class Discovered(Arrow):

    """
    This kind of arrow indicates that a key handle discovered a sub-key throught enumeration.
    """

    source : Handle
    destination : Key





class QueriesEntry(Arrow):

    """
    This kind of arrow indicates that a key entry was queried by a key handle. 
    """

    source : KeyEntry
    destination : Handle





class SetsEntry(Arrow):

    """
    This kind of arrow indicates that a key handle set the value of key entry.
    """

    source : Handle
    destination : KeyEntry





class DeletesEntry(Edge):

    """
    This kind of edge indicates that a key handle deleted a key entry.
    """

    source : Handle
    destination : KeyEntry





class ChangesTowards(Arrow):

    """
    This kind of arrow indicates that a key entry's content changed to another.
    """

    label : str = ""

    source : KeyEntry
    destination : KeyEntry





class ReferencesFileSystem(Edge):
    
    """
    This kind of edge indicates that a key's entry references a file or a folder.
    """

    source : KeyEntry
    destination : File | Directory





del Arrow, Directory, Edge, File, Handle, Host, Key, KeyEntry, Process, logger


"""
This module contains integration protocols for this behavioral package.
"""

from ....logger import logger
from ...build import BuildingPhase
from ...utils import chrono
from ..filesystem.entities import File, Handle
from ..network.entities import Connection, Socket
from . import entities, relations

__all__ = ["register_read_operation", "register_write_operation"]





logger.info("Loading integrations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

@chrono
def register_read_operation(target : File | Connection, vector : Handle | Socket, content : bytes | bytearray | memoryview | str, offset : int | None = None):
    """
    Registers a read operation to be added to the corresponding DiffNodes.
    """
    from .entities import Diff
    from .relations import IsDiffOf
    from .utils import Read
    if isinstance(content, str):
        content = content.encode("utf-8")
    if target not in Diff._active_target_diff:
        t = Diff()
        IsDiffOf(t, target)
        Diff._active_target_diff[target] = t
    else:
        t = Diff._active_target_diff[target]
    if vector not in Diff._active_vector_diff:
        v = Diff()
        IsDiffOf(v, vector)
        Diff._active_vector_diff[vector] = v
    else:
        v = Diff._active_vector_diff[vector]
    
    if offset == None:
        t.add_operation(Read(t.last_pos(), content))
        v.add_operation(Read(v.last_pos(), content))
    else:
        t.add_operation(Read(offset, content))
        v.add_operation(Read(offset, content))

@chrono
def register_write_operation(target : File | Connection, vector : Handle | Socket, content : bytes | bytearray | memoryview | str, offset : int | None = None):
    """
    Registers a write operation to be added to the corresponding DiffNodes.
    """
    from .entities import Diff
    from .relations import IsDiffOf
    from .utils import Write
    if isinstance(content, str):
        content = content.encode("utf-8")
    if target not in Diff._active_target_diff:
        t = Diff()
        IsDiffOf(t, target)
        Diff._active_target_diff[target] = t
    else:
        t = Diff._active_target_diff[target]
    if vector not in Diff._active_vector_diff:
        v = Diff()
        IsDiffOf(v, vector)
        Diff._active_vector_diff[vector] = v
    else:
        v = Diff._active_vector_diff[vector]
    
    if offset == None:
        t.add_operation(Write(t.last_pos(), content))
        v.add_operation(Write(v.last_pos(), content))
    else:
        t.add_operation(Write(offset, content))
        v.add_operation(Write(offset, content))

__N_diff_comparison_phase = BuildingPhase.request_finalizing_phase()
__N_diff_normalization_phase = BuildingPhase.request_finalizing_phase()
__N_diff_fusion_phase = BuildingPhase.request_finalizing_phase()
        

@chrono
def finalize_diff_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all Diff nodes to compute their data attributes.
    """
    from time import time_ns

    from Viper.format import duration

    from ....logger import logger
    from ...config import CompilationParameters
    from .entities import Diff
    if e.major == "Finalizer" and e.minor == __N_diff_comparison_phase:
        logger.debug("Finalizing{} {} Diff nodes.".format(" and comparing" if not CompilationParameters.SkipLevenshteinForDiffNodes else "",len(Diff)))
        n = len(Diff)
        t0 = time_ns()
        t = t0
        for i, d in enumerate(Diff):
            d.compute_data()
            if (time_ns() - t) / 1000000000 > 15:
                t = time_ns()
                logger.debug("Finalizing {} Diff nodes : {:.2f}%. ETA : {}".format(len(Diff), (i + 1) / n * 100, duration(round((t - t0) / (i + 1) * (n - i - 1)))))

@chrono
def normalize_diff_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all Diff nodes to compare their diff file sizes and change their size accordingly.
    """
    from typing import Type

    from ....logger import logger
    from ...colors import Color
    from ...graph import Vertex
    from ..filesystem import File, Handle
    from ..network import Connection, Socket
    from .entities import Diff

    def project_range(x : float, sa : float, sb : float, da : float, db : float) -> float:
        if sa == sb:
            return (da + db) / 2
        p = (x - sa) / (sb - sa)
        return p * (db - da) + da

    def normalize_neighbor(cls : Type[Vertex]):
        if len(cls) > 0:
            logger.debug("Normalizing {} {} nodes.".format(len(cls), cls.__name__))
            a, b = min(sum(v.size for v in u.neighbors() if isinstance(v, Diff)) for u in cls), max(sum(v.size for v in u.neighbors() if isinstance(v, Diff)) for u in cls)
            minsize = Diff.min_size
            maxsize = Diff.max_size
            for u in cls:
                u.size = project_range(sum(v.size for v in u.neighbors() if isinstance(v, Diff)), a, b, minsize, maxsize)

    if e.major == "Finalizer" and e.minor == __N_diff_normalization_phase:
        logger.debug("Normalizing sizes of {} Diff nodes.".format(len(Diff)))
        if len(Diff) == 0:
            return
        a, b = min(len(d.glob) for d in Diff), max(len(d.glob) for d in Diff)
        minsize = 0.5
        maxsize = 3.0
        for d in Diff:
            d.size = project_range(len(d.glob), a, b, minsize, maxsize)
    
        for cls in (File, Handle, Socket, Connection):
            normalize_neighbor(cls)
        
        logger.debug("Normalizing colors of {} Diff nodes.".format(len(Diff)))
        a, b = min(d.glob_entropy for d in Diff), max(d.glob_entropy for d in Diff)
        for d in Diff:
            d.color = Color.linear((Diff.diff_low_entropy_color, Diff.diff_high_entropy_color), (1 - d.glob_entropy / 8, d.glob_entropy / 8))

@chrono
def fuse_diff_nodes(e : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause all identical Diff nodes to be merged with their size increasing.
    """
    from ....logger import logger
    from ...graph import Graph
    from .entities import Diff

    groups : dict[Diff, int] = {}

    def merge(u : Diff, v : Diff):
        """
        Merges vertex v into vertex u (v gets deleted at the end).
        """
        for e in v.edges:
            if e.source is v:
                e.source = u
            if e.destination is v:
                e.destination = u
        v.edges.clear()
        for g in Graph.active_graphs():
            g.remove(v)
        groups[u] += 1

    if e.major == "Finalizer" and e.minor == __N_diff_fusion_phase:
        logger.debug("Fusing {} Diff nodes.".format(len(Diff)))
        for u in Diff:
            merged = False
            for v in filter(lambda v : v in groups, Diff):
                if u.content == v.content:
                    merge(v, u)
                    merged = True
                    break
            if not merged:
                groups[u] = 1
        
        resize_work = [u for u, n in groups.items() if n > 1]
        if resize_work:
            logger.debug("Resizing {} fused Diff nodes. Actually lost {} Diff nodes.".format(len(resize_work), len(Diff) - len(groups)))
            for u in resize_work:
                u.size *= groups[u] ** 0.5

        
BuildingPhase.add_callback(finalize_diff_nodes)
BuildingPhase.add_callback(normalize_diff_nodes)
BuildingPhase.add_callback(fuse_diff_nodes)





del BuildingPhase, Connection, File, Handle, Socket, chrono, entities, finalize_diff_nodes, fuse_diff_nodes, logger, normalize_diff_nodes, relations


"""
This module contains Edge and Arrow subclasses for this behavioral package.
"""

from ....logger import logger
from ...config import SwitchSetting
from ...colors import Color
from ...graph import Arrow, Edge, Vertex
from ..data.entities import Data
from ..execution.entities import Call, Process
from .entities import Connection, Host, Socket

__all__ = ["SpawnedProcess", "HasSocket", "HasConnection", "Communicates", "CreatesSocket", "Binds", "Connects", "Sends", "Receives", "Conveys", "Closes", "CloseSocket", "ListensOn", "Shutdown", "Accepts", "Duplicates"]





logger.info("Loading relations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class SpawnedProcess(Edge):
    
    """
    This kind of edge indicates that a machine hosts a process.
    """

    source : Host
    destination : Process





class HasSocket(Edge):

    """
    This kind of edge indicates that a process opened a socket.
    """

    label : str = ""

    source : Process
    destination : Socket





class HasConnection(Edge):

    """
    This kind of edge indicates that a socket makes a connection.
    """

    label : str = ""

    source : Socket
    destination : Connection





class Communicates(Edge):

    """
    This kind of edge indicates that two hosts communicate via a connection
    """

    source : Connection
    destination : Host





class CreatesSocket(Edge):

    """
    This kind of edge indicates that a system call created a socket.
    """

    label : str = ""

    source : Call
    destination : Socket





class Binds(Edge):

    """
    This kind of edge indicates that a system call bound a connection to a local address.
    """

    __slots__ = {
        "src" : "The source (local) address of the connection",
    }

    __pickle_slots__ = {
        "src"
    }

    label : str = ""

    source : Call
    destination : Connection

    def __init__(self, source: Vertex, destination: Vertex, *, auto_write: bool = True) -> None:
        from typing import Any
        super().__init__(source, destination, auto_write=auto_write)
        self.src : Any = None





class Connects(Edge):

    """
    This kind of edge indicates that a system call connected a connection to a remote address.
    """

    __slots__ = {
        "dest" : "The destination (remote) address of the connection"
    }

    __pickle_slots__ = {
        "dest"
    }

    label : str = ""

    source : Call
    destination : Connection

    def __init__(self, source: Vertex, destination: Vertex, *, auto_write: bool = True) -> None:
        from typing import Any
        super().__init__(source, destination, auto_write=auto_write)
        self.dest : Any = None





class Sends(Arrow):

    """
    This kind of arrow indicates that a system call sent data through a connection.
    """

    source : Call
    destination : Data





class Receives(Arrow):

    """
    This kind of arrow indicates that a system call received data through a connection.
    """

    source : Data
    destination : Call





class Conveys(Arrow):
    
    """
    This kind of arrow indicates that a connection conveyed a message.
    """

    label : str = ""

    source : Connection | Data
    destination : Connection | Data





class Closes(Edge):

    """
    This kind of edge indicates that a system call closed a connection.
    """
    
    label : str = ""

    source : Call
    destination : Connection





class CloseSocket(Edge):

    """
    This kind of edge indicates that a system call closed a socket object.
    """

    label : str = ""

    source : Call
    destination : Socket





class ListensOn(Edge):

    """
    This kind of edge indicates that a system call set a socket to listening mode.
    """

    source : Call
    destination : Socket





class Shutdown(Edge):

    """
    This kind of edge indicates that a connection was shutdown by a system call.
    """

    label : str = ""

    source : Call
    destination : Connection





class Accepts(Edge):

    """
    This kind of arrow indicates that a system call accepted a connection from a remote address.
    """

    __slots__ = {
        "dest" : "The destination (remote) address of the connection"
    }

    __pickle_slots__ = {
        "dest"
    }

    source : Call
    destination : Connection

    def __init__(self, source: Vertex, destination: Vertex, *, auto_write: bool = True) -> None:
        from typing import Any
        super().__init__(source, destination, auto_write=auto_write)
        self.dest : Any = None





class Duplicates(Arrow):

    """
    This kind of arrow indicates that a socket was duplicated to form a similar socket (example: after a call to accept()).
    """

    source : Socket
    destination : Socket





del Arrow, Call, Color, SwitchSetting, Connection, Data, Edge, Host, Process, Socket, Vertex, logger


"""
This is the MetaLib! It is a library for MetaGraphs, a pattern storage.

All existing MetaGraphs are loaded with the package.
To create new MetaGraphs, use the module croutons.metalib.interactive.
"""

from ...bakery.logger import logger
from .utils import entries, load

logger.info("Loading MetaGraph library.")

__all__ = []





g = globals()
for name in entries():
    try:
        g[name] = load(name)
    except:
        from traceback import print_exc
        print(f"Error while loading MetaGraph '{name}':")
        print_exc()
        del print_exc
        continue
    __all__.append(name)
    del name





del logger, load, entries, g


"""
This module contains Edge and Arrow subclasses for this behavioral package.
"""

from ....logger import logger
from ...config import SwitchSetting
from ...colors import Color
from ...graph import Arrow, Edge, Vertex
from ..filesystem.entities import Directory, File
from .entities import Call, Process, Thread

__all__ = ["HasChildProcess", "UsesAsArgument", "Runs", "HasThread", "HasFirstCall", "FollowedBy", "NextSignificantCall", "StartedThread", "InjectedThread", "StartedProcess"]





logger.info("Loading relations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class HasChildProcess(Arrow):

    """
    This kind of arrow indicates that a process created another one.
    """

    source : Process
    destination : Process





class UsesAsArgument(Edge):

    """
    This kind of edge indicates that a process used a file or directory in its command line.
    """

    source : Process
    destination : File | Directory




class Runs(Edge):
    
    """
    This kind of edge indicates process ran a file as its program.
    """

    source : Process
    destination : File





class HasThread(Edge):

    """
    This kind of edge indicates that a process hosts a thread.
    """

    label : str = ""

    source : Process
    destination : Thread





class HasFirstCall(Edge):

    """
    This kind of edge indicates what was the first system call of a thread.
    """

    label : str = ""

    source : Thread
    destination : Call





class FollowedBy(Arrow):

    """
    This kind of arrow indicates that a system call was followed by another one.
    """

    label : str = ""

    source : Call
    destination : Call





class NextSignificantCall(Arrow):
    
    """
    This kind of arrows links two Call nodes that happened one after the other in the same thread both nodes are linked to other types of nodes.
    """

    label : str = ""

    source : Call
    destination : Call





class StartedThread(Edge):

    """
    This kind of edge indicates that a system call started a new thread.
    """

    __slots__ = {
        "remote" : "Indicates if the started thread was started in another process (remote thread)"
    }

    __pickle_slots__ = {
        "remote"
    }

    label : str = ""

    source : Call
    destination : Thread

    def __init__(self, source: Vertex, destination: Vertex, *, auto_write: bool = True) -> None:
        super().__init__(source, destination, auto_write=auto_write)
        self.remote : bool = False
    




class InjectedThread(Arrow):

    """
    This kind of arrow indicates a process created a remote thread.
    """

    source : Process
    destination : Thread





class StartedProcess(Edge):

    """
    This kind of edge indicates that a system call started a new process.
    """

    label : str = ""

    source : Call
    destination : Process





del Arrow, Call, Color, SwitchSetting, Directory, Edge, File, Process, Thread, Vertex, logger


"""
This module contains configuration objects for execution graphs.
"""

from abc import abstractmethod
from typing import Any, Generic, Self, TypeVar
from .colors import Color
from Viper.meta.decorators import hybridmethod, staticproperty

__all__ = ["Setting", "ColorSetting", "SizeSetting", "SwitchSetting", "WeightSetting", "CompilationParameters", "ajust_for_background_color"]





T = TypeVar("T")

class Setting(Generic[T]):

    """
    Just a class to hold settings for BAGUETTE. Cannot be used as such, you need to specialize it as a generic class of the type of object the setting will hold:

    >>> s = Setting(Color.white)
    ValueError: Cannot instanciate unspecialized Setting class
    >>> ColorMap = Setting[Color]
    >>> s = ColorMap(Color.white)
    >>> class A:
    ...     x = ColorMap(Color.red)
    ... 
    >>> a = A()
    >>> a.x
    Color.red
    >>> a.x = 1
    TypeError: Expected Color, got 'int'
    >>> a.x = Color.yellow
    >>> a.x
    Color.yellow
    >>> A.x         # Note that settings are actually affected at a class level.
    Color.yellow
    >>> del a.x
    >>> a.x
    Color.red
    
    The advantage of Settings over simple properties is that references are hold at the class level to all the settings declared in all class namespaces:

    >>> Setting.list()
    {(<class '__main__.A'>, 'x'): <baguette.bakery.source.config.Setting[Color] object at 0x0000016345399590>}
    >>> IntMap = Setting[int]
    >>> IntMap.list()
    {}
    >>> ColorMap.list()
    {(<class '__main__.A'>, 'x'): <baguette.bakery.source.config.Setting[Color] object at 0x0000016345399590>}
    >>> a.x = Color.yellow
    >>> Setting.list()[A, "x"].value = Color.black
    >>> a.x
    Color.black
    """

    __slots__ = {
        "__default" : "The default value of this Setting",
        "__current" : "The current value set for this Setting",
    }

    __per_class : dict[tuple[type, str], "Setting"] = {}

    @classmethod
    def __class_getitem__(cls, key : type) -> type["Setting"]:
        if not isinstance(key, type):
            raise TypeError(f"Expected type, got '{type(key).__name__}'")
        from Viper.meta.decorators import staticproperty
        return type(cls)(f"Setting[{key.__name__}]", (Setting, ), {"cls" : staticproperty(lambda : key)})

    def __new__(cls, default : T) -> Self:
        if not issubclass(cls, Setting):
            raise TypeError("Cannot use Setting.__new__ on something that is not a subclass of Setting")
        res : Setting = super().__new__(cls)
        try:
            res.cls
        except NotImplementedError:
            raise ValueError("Cannot instanciate unspecialized Setting class") from None
        return res

    def __init__(self, default : T) -> None:
        self.__default = default
        self.__current = default
        if not isinstance(default, self.cls):
            raise TypeError(f"Expected {self.cls.__name__}, got '{type(default).__name__}'")

    @staticproperty
    @abstractmethod
    def cls() -> type[T]:
        """
        The type of value expected for this Setting.
        """
        raise NotImplementedError

    @property
    def value(self) -> T:
        """
        The value that this Setting currently points to.
        """
        return self.__current
    
    @value.setter
    def value(self, value : T):
        if not isinstance(value, self.cls):
            raise TypeError(f"Expected {self.cls.__name__}, got '{type(value).__name__}'")
        self.__current = value
    
    @value.deleter
    def value(self):
        self.__current = self.__default

    @classmethod
    def list(cls, namespace : type | None = None) -> dict[tuple[type, str], "Setting[T]"]:
        """
        Lists all the available settings in a dict object.
        Keys are tuples in the form (<class>, <name>) for each Vertex or Edge subclass with the name of the attribute which holds the Setting.
        Values are the Setting objects themselves.
        A filter class can be given to filter only the settings present in this class and its subclasses.
        """
        if namespace is not None and not isinstance(namespace, type):
            raise TypeError(f"Expected type or None, got '{type(namespace).__name__}")
        if namespace:
            if cls is Setting:
                return {(c, n) : s for (c, n), s in Setting.__per_class.items() if issubclass(c, namespace)}
            else:
                return {(c, n) : s for (c, n), s in Setting.__per_class.items() if s.cls is cls.cls and issubclass(c, namespace)}
        else:
            if cls is Setting:
                return {(c, n) : s for (c, n), s in Setting.__per_class.items()}
            else:
                return {(c, n) : s for (c, n), s in Setting.__per_class.items() if s.cls is cls.cls}
    
    def specialize_subclass(self, cls : type) -> Self:
        """
        Copies the setting to a subclass, allowing the subclass to have a different value for this setting.
        Returns the newly created setting.
        """
        if not isinstance(cls, type):
            raise TypeError(f"Expected type, got '{type(cls).__name__}'")
        target_cls = None
        name = None
        for (c, n), s in Setting.__per_class.items():
            if s is self:
                target_cls = c
                name = n
                break
        if target_cls is None or name is None:
            raise ValueError("Cannot specialize anonymous setting (This setting is not in the code of any known class)")
        if not issubclass(cls, target_cls):
            raise ValueError("Expected subclass of the class this setting is registered in.")
        cp = type(self)(self.__default)
        setattr(cls, name, cp)
        cp.__set_name__(cls, name)
        return cp

    def __set_name__(self, owner : type, name : str):
        from ..logger import logger
        logger.debug(f"New '{self.cls.__name__}' setting registered for class '{owner.__name__}' with name '{name}'.")
        Setting.__per_class[(owner, name)] = self

    def __get__(self, instance : Any, owner : type | None = None) -> T:
        return self.__current
    
    def __set__(self, instance : Any, value : T):
        if not isinstance(value, self.cls):
            raise TypeError(f"Expected {self.cls.__name__}, got '{type(value).__name__}'")
        self.__current = value

    def __delete__(self, instance : Any):
        self.__current = self.__default





ColorSetting = Setting[Color]
SizeSetting = Setting[float]
WeightSetting = Setting[float]
SwitchSetting = Setting[bool]
FlagSetting = Setting[bool]





def ajust_for_background_color(background_color : Color, tolerance : float = 0.15):
    """
    This function will change the values of the Color Settings to ajust those that would not be visible for a background of the given color.
    """
    from ..logger import logger

    for (cls, name), color_setting in ColorSetting.list().items():
        if sum((a - b) ** 2 for a, b in zip(color_setting.value, background_color)) ** 0.5 < tolerance:
            logger.info(f"Changing color setting '{cls.__name__}.{name}'.")
            color_setting.value = color_setting.value.negative_to(background_color)






class __CompilationParameters:

    """
    This class holds some compilations parameters that some modules may want to use.
    """

    SkipLevenshteinForDataNodes = FlagSetting(False)
    SkipLevenshteinForDiffNodes = FlagSetting(False)

CompilationParameters = __CompilationParameters()





del abstractmethod, Any, Generic, Self, TypeVar, Color, hybridmethod, staticproperty        # type: ignore


"""
This module adds some useful functions and classes for the compilation of BAGUETTE graphs.
"""

import pathlib
from functools import cache
from typing import (Any, Callable, Generic, Iterable, Iterator, ParamSpec, Set,
                    TypeVar)

from Viper.debugging.chrono import Chrono

from .graph import Edge, Vertex

__all__ = ["chrono", "extract_pickle_slots", "path_factory", "is_path", "parse_command_line", "System"]





chrono = Chrono()
chrono.enabled = False





@cache
def extract_pickle_slots(o : type[Vertex | Edge]) -> Set[str]:
    s : set[str]
    if not hasattr(o, "__pickle_slots__"):
        s = set()
    else:
        s = set(o.__pickle_slots__) # type: ignore
    for b in s:
        if not isinstance(b, str):
            raise RuntimeError("__pickle_slots__ should only contain str, got a " + repr(type(b).__name__))
    return s.union(*[extract_pickle_slots(b) for b in o.__bases__])


def path_factory(path : str) -> pathlib.PurePath:
    """
    Returns the correct path factory class for the given graph (Either WindowsPurePath or PosixPurePath).
    """
    from pathlib import PurePosixPath, PureWindowsPath

    from .graph import Graph
    platform = Graph.active_graphs()[-1].data["platform"].lower() if Graph.active_graphs() else "windows"
    if "windows" in platform:
        return PureWindowsPath(path)
    else:
        return PurePosixPath(path)


def is_path(s : str) -> bool:
    """
    Returns True if the given string represents a path for the current platform.
    """
    from pathlib import PurePosixPath, PureWindowsPath

    from .graph import Graph
    platform = Graph.active_graphs()[-1].data["platform"].lower() if Graph.active_graphs() else "windows"
    if "windows" in platform:
        forbidden = set("<>\"|?*")
        if forbidden.intersection(s):           # Contains characters that are forbidden in Windows paths.
            return False
        if s.startswith("-") or s.startswith("/"):  # Could be a path, but it is more likely to be flag
            return False
        try:
            if PureWindowsPath(s).is_reserved():
                return False
            return True
        except:
            return False
    else:
        if s.startswith("-"):                       # Could be a path, but it is more likely to be flag
            return False
        try:
            if PurePosixPath(s).is_reserved():
                return False
            return True
        except:
            return False
        

def parse_command_line(s : str) -> list[str]:
    """
    Parses command line with the syntax of the current platform's shell. Returns the argument vector.
    """
    import re

    from .graph import Graph

    platform = Graph.active_graphs()[-1].data["platform"].lower() if Graph.active_graphs() else "windows"
    if "windows" not in platform:
        RE_CMD_LEX = r'''"((?:\\["\\]|[^"])*)"|'([^']*)'|(\\.)|(&&?|\|\|?|\d?\>|[<])|([^\s'"\\&|<>]+)|(\s+)|(.)'''
    else:
        RE_CMD_LEX = r'''"((?:""|\\["\\]|[^"])*)"?()|(\\\\(?=\\*")|\\")|(&&?|\|\|?|\d?>|[<])|([^\s"&|<>]+)|(\s+)|(.)'''

    args : list[str] = []
    accu = None   # collects pieces of one arg
    for qs, qss, esc, pipe, word, white, fail in re.findall(RE_CMD_LEX, s):
        if word:
            pass   # most frequent
        elif esc:
            word = esc[1]
        elif white or pipe:
            if accu is not None:
                args.append(accu)
            if pipe:
                args.append(pipe)
            accu = None
            continue
        elif fail:
            raise ValueError("invalid or incomplete shell string")
        elif qs:
            word = qs.replace('\\"', '"').replace('\\\\', '\\')
            if "windows" in platform:
                word = word.replace('""', '"')
        else:
            word = qss   # may be even empty; must be last

        accu = (accu or '') + word

    if accu is not None:
        args.append(accu)

    return args


P = TypeVar("P")
P2 = TypeVar("P2")
CB_Args = ParamSpec("CB_Args")

class System(Generic[P]):

    def __init__(self) -> None:
        from typing import Callable
        self.__system : set[P] = set()
        self.__population : set[P] = set()
        self.__at_collapse : list[tuple[Callable[..., None], tuple, dict[str, Any]]] = []
        collapse = self.__at_collapse
        population = self.__population
        system = self.__system

        class Particule(Generic[P2]):

            __slots__ = {"__particule"}

            def __init__(self, particule : P2) -> None:
                self.__particule = particule

            def __enter__(self):
                if self.__particule in system:
                    raise RuntimeError("Trying to entangle Particule with itself : cannot copy quantum information.")
                population.discard(self.__particule)    # type: ignore
                system.add(self.__particule)            # type: ignore
           
            def __exit__(self, exc_type, exc_value, traceback):
                if self.__particule not in system:
                    raise RuntimeError("Trying to collapse non-quantum object.")
                system.remove(self.__particule)         # type: ignore
                if not system:  # The system has fully collapsed : time for callbacks
                    for cb, args, kwargs in collapse:
                        cb(*args, **kwargs)
                    collapse.clear()
        
        self.Entangled : type[Particule[P]] = Particule

    def __contains__(self, p : P) -> bool:
        return p in self.__system
    
    def __len__(self) -> int:
        return len(self.__system)
    
    @property
    def population(self) -> set[P]:
        """
        A set of elements that can be included in the system.
        """
        return self.__population
    
    @population.setter
    def population(self, v : Iterable[P]):
        from typing import Iterable
        if not isinstance(v, Iterable):
            raise TypeError("Expected iterable, got " + repr(type(v).__name__))
        self.__population.clear()
        self.__population.update(k for k in v if k not in self)

    def include(self) -> Iterator[P]:
        """
        Yields all successives items of the population to be added to the system.
        """
        while self.__population:
            yield self.__population.pop()
    
    def add_callback(self, cb : Callable[CB_Args, None], *args : CB_Args.args, **kwargs : CB_Args.kwargs):
        """
        Adds a callable to be called with given arguments when the system completely resolves (i.e. when all entangled wave functions have collapsed).
        """
        if not self.__system:
            raise RuntimeError("Cannot program a callback before the experiment has started. Add at least one particle first.")
        self.__at_collapse.append((cb, args, kwargs))
    
    def remove_callback(self, cb : Callable[CB_Args, None]):
        """
        Removes all callbacks to the given function.
        """
        to_pop = []
        for i, (cbi, argsi, kwargsi) in enumerate(self.__at_collapse):
            if cb == cbi:
                to_pop.append(i)
        for i in reversed(to_pop):
            self.__at_collapse.pop(i)





del pathlib, cache, Any, Callable, Iterable, Iterator, ParamSpec, Set, TypeVar, Chrono, Edge, Vertex, CB_Args


"""
This module contains integration protocols for this behavioral package.
"""

from ....logger import logger
from ...utils import chrono
from ..filesystem.integration import NewFile
from . import entities, relations

__all__ = []





logger.info("Loading integrations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

@chrono
def link(e : NewFile):
    from ...graph import find_or_create
    from .entities import Import
    from .relations import IsFile
    if e.file.path.suffix.lower().endswith(".dll"):
        I, ok = find_or_create(Import, path = e.file.name)
        if not ok:
            import re
            expr1 = re.compile(r"([a-zA-Z0-9_.]+)(\..*?)$")
            match = expr1.search(e.file.name)
            if match:
                I.name = match.group(1).lower()
            else:
                I.name = e.file.name
        IsFile(I, e.file)

NewFile.add_callback(link)





del NewFile, chrono, entities, link, logger, relations


"""
This module defines all graph-related objects.
"""

from threading import RLock, Thread
from typing import Any, Callable, Iterable, Iterator, Optional, TypeVar, Union
from .colors import Color
from .config import ColorSetting, SizeSetting, SwitchSetting, WeightSetting
from Viper.meta.iterable import InstanceReferencingClass

__all__ = ["Vertex", "UniqueVertex", "Edge", "Arrow", "Graph", "find_or_create"]





class Vertex(metaclass = InstanceReferencingClass):

    """
    A vertex for a graph. Can be linked to other vertices and added to a graph.
    """

    __slots__ = {
        "edges" : "The set of edges linking this vertex to others.",
        "parent" : "A vertex that 'contains' this one.",
        "children" : "The tuple of vertices (directly) contained by this one.",
        "graph" : "The graph object containing this vertex",
        "__size" : "The customized size of the vertex",
        "__color" : "The customized color of the vertex"
        }
    
    __pickle_slots__ = {
        "color",
        "parent",
        "size"
    }

    default_color = ColorSetting(Color.white)
    default_size = SizeSetting(2.0)

    def __init__(self, *, parent : Optional["Vertex"] = None) -> None:
        from .colors import Color
        if parent != None and not isinstance(parent, Vertex):
            raise TypeError("Expected vertex for parent, got " + repr(parent.__class__.__name__))
        self.edges : set[Edge] = set()
        self.__color : Color | None = None
        self.__size : float | None = None
        self.parent : Vertex | None = parent
        self.graph : Graph | None = None
        if parent:
            parent.children += (self, )
        self.children : tuple[Vertex, ...] = ()
        for g in Graph.active_graphs():
            g.vertices.add(self)

    __vertexing = False
    __comparing : set[tuple["Vertex", ...]] = set()

    @property
    def color(self) -> Color:
        """
        The Color of this Vertex.
        """
        if self.__color is not None:
            return self.__color
        return self.default_color
    
    @color.setter
    def color(self, value : Color):
        from .colors import Color
        if not isinstance(value, Color):
            raise TypeError(f"Expected Color, got '{type(value).__name__}'")
        self.__color = value

    @color.deleter
    def color(self):
        self.__color = None

    @property
    def size(self) -> float:
        """
        The size of this Vertex.
        """
        if self.__size is not None:
            return self.__size
        return self.default_size
    
    @size.setter
    def size(self, value : float):
        if not isinstance(value, float):
            try:
                value = float(value)
            except:
                pass
        if not isinstance(value, float):
            raise TypeError("Expected float, got " + repr(type(value).__name__))
        if value < 0 or value in (float("inf"), float("nan")):
            raise ValueError("Expected positive finite number for size, got " + repr(value))
        self.__size = value
    
    @size.deleter
    def size(self):
        self.__size = None

    @property
    def label(self) -> str:
        """
        The label used when plotting this vertex.
        """
        return type(self).__name__

    def __repr__(self) -> str:
        """
        Implements repr(self)
        """
        if Vertex.__vertexing:
            return str(type(self)) + "()"
        Vertex.__vertexing = True
        r = type(self).__name__ + "(" + ", ".join(str(name) + " = " + str(getattr(self, name)) for name in self.__slots__ if not name.startswith("_")) + ")"
        Vertex.__vertexing = False
        return r
    
    def __str__(self) -> str:
        """
        Implements str(self)
        """
        if Vertex.__vertexing:
            return str(type(self))
        Vertex.__vertexing = True
        r = "(" + ", ".join(str(name) + " = " + str(getattr(self, name)) for name in self.__slots__ if not name.startswith("_")) + ")"
        Vertex.__vertexing = False
        return r
    
    def __hash__(self) -> int:
        """
        Implements hash(self)
        """
        n = 0
        for name in dir(self):
            if not name.startswith("__") and not name.endswith("__") and name not in Vertex.__slots__:
                try:
                    n += hash(name) * hash(getattr(self, name))
                except:
                    pass
        return n
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o
        """
        if self is o:
            return True
        if (self, o) in self.__comparing or (o, self) in self.__comparing:
            return True
        if type(self) != type(o):
            return False
        self.__comparing.add((self, o))
        res = all(getattr(self, name) == getattr(o, name) for name in type(self).__slots__)
        self.__comparing.remove((self, o))
        return res
    
    def __reduce__(self) -> str | tuple[Any, ...]:
        """
        Implements dumping of self
        """
        # Note : When serializing a vertex, the edges/arrows should not be serialized. These will be handled independently when serializing a graph.
        pickle_data = super().__reduce__()
        if isinstance(pickle_data, str):
            raise TypeError("Expected tuple from parent's __reduce__, got str")
        func, args, state = pickle_data
        state : dict
        if "edges" in state or "graph" in state:
            raise RuntimeError("Cannot pickle edges or graph of a vertex")
        return type(self), (), state
    
    def __setstate__(self, state : dict[str, Any]):
        """
        Implements loading of self
        """
        for k, v in state.items():
            setattr(self, k, v)
    
    def __getstate__(self) -> dict[str, Any]:
        """
        Implements dumping of self
        """
        from .utils import extract_pickle_slots
        return {name : getattr(self, name) for name in extract_pickle_slots(type(self))}
    
    def __copy__(self) -> "Vertex":
        """
        Implements copy of self
        """
        from inspect import ismethod
        cp = type(self)()
        cp.color = self.color
        for name in dir(self):
            if not name.startswith("_") and not name.endswith("_") and name not in Vertex.__slots__ and not ismethod(getattr(self, name)):
                setattr(cp, name, getattr(self, name))
        return cp
    
    def __deepcopy__(self, memo : dict[int, Any]) -> "Vertex":
        """
        Implements deepcopy of self
        """
        from copy import deepcopy
        from inspect import ismethod
        cp = type(self)()
        cp.color = self.color
        memo[id(self)] = cp
        for name in dir(self):
            if not name.startswith("_") and not name.endswith("_") and name not in Vertex.__slots__ and not ismethod(getattr(self, name)):
                setattr(cp, name, deepcopy(getattr(self, name), memo))
        return cp
                    
    def neighbors(self) -> Iterator["Vertex"]:
        """
        Iterates over all the neighbor vertices.
        """
        for e in self.edges:
            if e.source is self:
                yield e.destination
            else:
                yield e.source
    
    def outwards(self) -> Iterator["Vertex"]:
        """
        Iterates over the outwards neighbors of this vertex (neighbors linked by an outgoing arrow).
        """
        for e in self.edges:
            if isinstance(e, Arrow) and e.source is self:
                yield e.destination
    
    def inwards(self) -> Iterator["Vertex"]:
        """
        Iterates over the inwards neighbors of this vertex (neighbors linked by an incomming arrow).
        """
        for e in self.edges:
            if isinstance(e, Arrow) and e.destination is self:
                yield e.source
    
    def linked(self) -> Iterator["Vertex"]:
        """
        Iterates over the undirected neighbors if this vertex (neighbors linked by a strict edge).
        """
        for e in self.edges:
            if not isinstance(e, Arrow):
                yield (e.source if e is not e.source else e.destination)

    def connect(self, o : "Vertex", *, directional : bool = False) -> None:
        """
        Links this vertex to another. Directional indicates if the link should be an arrow instead of an edge.
        """
        if not isinstance(o, Vertex):
            raise TypeError("Expected vertex, got " + repr(o.__class__.__name__))
        if not isinstance(directional, bool):
            raise TypeError("Expected bool for directional, got" + repr(directional.__class__.__name__))
        if directional:
            e = Arrow(self, o)
        else:
            e = Edge(self, o)
        e.write()

    @classmethod
    def add_vertices_to_graph(cls : type["Vertex"], G : "Graph", fil : Optional[Callable[["Vertex"], bool]] = None):
        """
        Adds all vertices of this class to a graph.
        If given a filter function fil, only filtered vertices will be added.
        """
        if not isinstance(G, Graph):
            raise TypeError("Expected graph, got " + repr(G.__class__.__name__))
        if fil != None and not callable(fil):
            raise TypeError("Expected callable for filter, got " + repr(fil.__class__.__name__))
        if fil == None:
            G.vertices.update(cls)
        else:
            G.vertices.update(filter(fil, cls))
                




class UniqueVertex(Vertex):

    """
    Vertices from this class all have different hashes and always compare different (except when they are the same object).
    """

    def __hash__(self) -> int:
        """
        Implements hash(self).
        """
        return id(self)
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o.
        """
        return self is o
            




class Edge(metaclass = InstanceReferencingClass):

    """
    An (undirected) edge for a graph. Links two vertices together.
    """

    __slots__ = {
        "source" : "The source vertex.",
        "destination" : "The destination vertex.",
        "__weight" : "The customized weight of the edge",
        "__color" : "The customized color of the edge"
    }

    __pickle_slots__ = {
        "color",
        "source",
        "destination",
        "weight"
    }

    default_color = ColorSetting(Color.white)
    default_weight = WeightSetting(1.0)

    blend_vertices_colors = SwitchSetting(True)

    def __init__(self, source : Vertex, destination : Vertex, *, auto_write : bool = True) -> None:
        from .colors import Color
        if not isinstance(source, Vertex) or not isinstance(destination, Vertex):
            raise TypeError("Expected vertex, vertex, got " + repr(source.__class__.__name__) + " and " + repr(destination.__class__.__name__))
        if not isinstance(auto_write, bool):
            raise TypeError("Expected bool for write, got " + repr(auto_write.__class__.__name__))
        self.source : Vertex = source
        self.destination : Vertex = destination
        self.__color : Color | None = None
        self.__weight : float | None = None
        if auto_write:
            self.write()
        for g in Graph.active_graphs():
            g.edges.add(self)

    @property
    def color(self) -> Color:
        """
        The Color of this Vertex.
        """
        if self.__color is not None:
            return self.__color
        if self.blend_vertices_colors:
            from .colors import Color
            return Color.average(self.source.color, self.destination.color)
        else:
            return self.default_color
    
    @color.setter
    def color(self, value : Color):
        from .colors import Color
        if not isinstance(value, Color):
            raise TypeError(f"Expected Color, got '{type(value).__name__}'")
        self.__color = value

    @color.deleter
    def color(self):
        self.__color = None
        
    @property
    def weight(self) -> float:
        """
        The weight of this Edge.
        """
        if self.__weight is not None:
            return self.__weight
        return self.default_weight
    
    @weight.setter
    def weight(self, value : float):
        if not isinstance(value, float):
            try:
                value = float(value)
            except:
                pass
        if not isinstance(value, float):
            raise TypeError("Expected float for weight, got " + repr(type(value).__name__))
        if value < 0 or value in (float("inf"), float("nan")):
            raise ValueError("Expected positive finite number for weight, got " + repr(value))
        self.__weight = value

    @weight.deleter
    def weight(self):
        self.__weight = None

    @property
    def label(self) -> str:
        """
        The label used when plotting this edge.
        """
        return type(self).__name__
    
    def __repr__(self) -> str:
        """
        Implements repr(self)
        """
        return "edge(" + repr(self.source) + ", " + repr(self.destination) + ")"
    
    def __str__(self) -> str:
        """
        Implements str(self)
        """
        return str(self.source) + " -- " + str(self.destination)
    
    def __hash__(self) -> int:
        """
        Implements hash(self)
        """
        return hash(hash(self.source) + hash(type(self)) + hash(self.destination))
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o
        """
        if not isinstance(o, type(self)):
            return False
        return (self.source == o.source and self.destination == o.destination) or (self.source == o.destination and self.destination == o.source)
        
    def __setstate__(self, state : dict[str, Any]):
        """
        Implements loading of self
        """
        for k, v in state.items():
            setattr(self, k, v)
    
    def __getstate__(self) -> dict[str, Any]:
        """
        Implements dumping of self
        """
        from .utils import extract_pickle_slots
        return {name : getattr(self, name) for name in extract_pickle_slots(type(self))}
    
    def __copy__(self) -> "Edge":
        """
        Implements copy(self)
        """
        raise NotImplementedError
        
    def write(self):
        """
        Writes this edge in the edges sets of both vertices.
        """
        self.source.edges.add(self)
        self.destination.edges.add(self)
        if self.source.graph:
            self.source.graph.edges.add(self)
        elif self.destination.graph:
            self.destination.graph.edges.add(self)
    
    def delete(self) -> tuple[Vertex, Vertex]:
        """
        Deletes the link. (Deletes it from the vertices egde sets)
        """
        self.source.edges.discard(self)
        self.destination.edges.discard(self)
        if self.source.graph:
            self.source.graph.edges.discard(self)
        elif self.destination.graph:
            self.destination.graph.edges.discard(self)
        return self.source, self.destination
    
    @classmethod
    def add_edges_to_graph(cls : type["Edge"], G : "Graph", fil : Optional[Callable[["Edge"], bool]] = None):
        """
        Adds all edges of this class to a graph.
        If given a filter function fil, only filtered edges will be added.
        """
        if not isinstance(G, Graph):
            raise TypeError("Expected graph, got " + repr(G.__class__.__name__))
        if fil != None and not callable(fil):
            raise TypeError("Expected callable for filter, got " + repr(fil.__class__.__name__))
        if fil == None:
            G.edges.update(cls)
        else:
            G.edges.update(filter(fil, cls))

    
        


class Arrow(Edge):

    """
    An arrow for a graph. Links two vertices together.
    """
    
    def __repr__(self) -> str:
        """
        Implements repr(self).
        """
        return "arrow(" + repr(self.source) + ", " + repr(self.destination) + ")"
    
    def __str__(self) -> str:
        """
        Implements str(self).
        """
        return str(self.source) + " --> " + str(self.destination)
    
    def __hash__(self) -> int:
        """
        Implements hash(self).
        """
        return hash(-hash(self.source) + hash(type(self)) + hash(self.destination))
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o.
        """
        if not isinstance(o, type(self)):
            return False
        return self.source == o.source and self.destination == o.destination
    





CLS = TypeVar("CLS", bound=Vertex)

def find_or_create(cls : type[CLS], **kwargs) -> tuple[CLS, bool]:
    """
    Finds in the given class cls for a vertex which attributes have values matching the given keyword arguments.
    If there is one, returns it. Otherwise, creates it.
    Returns the object and a boolean indicating if the object existed before.
    """
    for ist in cls:
        if all(hasattr(ist, name) and getattr(ist, name) == value for name, value in kwargs.items()):
            return ist, True
    ist = cls()
    for name, value in kwargs.items():
        setattr(ist, name, value)
    return ist, False





class Graph:

    """
    A graph class. Add starting vertices and use explore() to complete the graph.
    Graph can be used in context managers to append all Vertices and Edges created in the current thread to the context Graph.
    """

    __slots__ = {
        "vertices" : "The set of vertices in this graph",
        "edges" : "The set of edges in this graph",
        "data" : "A dictionary holding additional data for this graph"
    }

    __active_graphs : dict[Thread, list["Graph"]] = {}

    def __init__(self) -> None:
        from typing import Any
        self.vertices : set[Vertex] = set()
        self.edges : set[Edge] = set()
        self.data : dict[str, Any] = {}
    
    def __enter__(self):
        """
        Implements with self.
        """
        from threading import current_thread
        if current_thread() not in Graph.__active_graphs:
            Graph.__active_graphs[current_thread()] = []
        Graph.__active_graphs[current_thread()].append(self)
    
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Implements with self.
        """
        from threading import current_thread
        Graph.__active_graphs[current_thread()].remove(self)

    def __eq__(self, o : object) -> bool:
        """
        Implements self == o. Equivalent to self is o.
        """
        return self is o
    
    @staticmethod
    def active_graphs() -> list["Graph"]:
        """
        Returns the list of all the active Graphs in the current thread.
        """
        from threading import current_thread
        if current_thread() not in Graph.__active_graphs:
            return []
        return Graph.__active_graphs[current_thread()].copy()

    def add_vertex(self, v : Vertex, explore : bool = True) -> None:
        """
        Adds a vertex to the graph and explores from this vertex.
        """
        if not isinstance(v, Vertex):
            raise TypeError("Expected vertex, got " + repr(v.__class__.__name__))
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if v not in self.vertices:
            self.vertices.add(v)
            v.graph = self
            if explore:
                self.explore(v)
    
    def explore(self, source : Optional[Vertex] = None) -> None:
        """
        Explores the graph for more linked vertices. If a vertex of the graph is given, only searches starting from that vertex. Otherwise, searches from all vertices.
        """
        if source == None:
            to_explore = self.vertices.copy()
        else:
            if not isinstance(source, Vertex):
                raise TypeError("Expected vertex, got " + repr(source.__class__.__name__))
            if source not in self.vertices:
                raise KeyError("Vertex not in graph.")
            to_explore = {source}
        from itertools import chain
        seen_vertices = set()
        seen_edges = set()
        while to_explore:
            u = to_explore.pop()
            seen_vertices.add(u)
            seen_edges.update(u.edges)
            u.graph = self
            next_vertices = chain(u.neighbors(), iter(u.children))
            if u.parent:
                next_vertices = chain(next_vertices, (u.parent, ))
            new = set()
            for v in next_vertices:
                if v not in seen_vertices:
                    new.add(v)
            to_explore.update(new)
        self.vertices.update(seen_vertices)
        self.edges.update(seen_edges)

    def pairs(self) -> Iterator[tuple[Vertex, Edge, Vertex]]:
        """
        Yields all (vertex u, edge e, vertex v) tuples where (u, v) is a pair of linked vertices. e may be an edge or an arrow directed from u to v.
        """
        for e in self.edges:
            yield e.source, e, e.destination

    def __len__(self) -> int:
        """
        Implements len(self). Returns the number of vertices.
        """
        return len(self.vertices)
    
    @property
    def n(self) -> int:
        """
        The number of vertices.
        """
        return len(self)
    
    @property
    def m(self) -> int:
        """
        The number of edges and arrows.
        """
        return len(self.edges)
        
    def append(self, value : Vertex | Edge, explore : bool = False):
        """
        Adds a vertex or an edge to the graph.
        If explore is True, explores the graph from the added vertex or the source vertex of the added edge.
        """
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if isinstance(value, Vertex):
            self.vertices.add(value)
            value.graph = self
            if explore:
                self.explore(value)
        elif isinstance(value, Edge):
            value.write()
            self.edges.add(value)
            if explore:
                self.explore(value.source)
        else:
            raise TypeError("Expected edge or vertex, got " + repr(value.__class__.__name__))
    
    def remove(self, value : Vertex | Edge):
        """
        Removes a vertex or an edge from the graph.
        When removing a vertex, it will also remove all edges/arrows connected to it.
        """
        if isinstance(value, Vertex):
            self.edges.difference_update(value.edges)
            for e in value.edges.copy(): 
                e.source.edges.discard(e)
                e.destination.edges.discard(e)
            self.vertices.discard(value)
        elif isinstance(value, Edge):
            self.edges.discard(value)
            value.source.edges.discard(value)
            value.destination.edges.discard(value)
        else:
            raise TypeError("Expected edge or vertex, got " + repr(type(value).__name__))

    def extend(self, values : Iterable[Vertex | Edge], explore : bool = False):
        """
        Extends the graph with an iterable of vertices and/or edges.
        """
        from typing import Iterable
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if not isinstance(values, Iterable):
            raise TypeError("Expected iterable, got " + repr(values.__class__.__name__))
        for value in values:
            if isinstance(value, Vertex):
                self.vertices.add(value)
                value.graph = self
                if explore:
                    self.explore(value)
            elif isinstance(value, Edge):
                value.write()
                self.edges.add(value)
                if explore:
                    self.explore(value.source)
            else:
                raise TypeError("Expected edge or vertex, got " + repr(value.__class__.__name__))
        
    def paint(self, c : Color):
        """
        Changes the colors of all vertices and edges/arrows of this graph to the given Color.
        """
        from .colors import Color
        if not isinstance(c, Color):
            raise TypeError("Expected Color, got " + repr(type(c).__name__))
        for u in self.vertices:
            u.color = c
        for e in self.edges:
            e.color = c

    def __getstate__(self) -> dict[str, Any]:
        """
        Implements dumping of self.
        """
        return {
            "vertices" : self.vertices,
            "edges" : self.edges, 
            "data" : self.data
        }

    def __setstate__(self, state : dict[str, Any]):
        """
        Implements loading of self.
        """
        for name, value in state.items():
            setattr(self, name, value)
        for v in self.vertices:
            v.graph = self
        for e in self.edges:
            e.write()
   
    # def __copy__(self) -> "graph":
    #     """
    #     Implements copy of self
    #     """
    #     from copy import copy
    #     translation = {u : copy(u) for u in self.vertices}
    #     cp = graph()
    #     cp.vertices.update(translation.values())
    #     for u, e, v in self.pairs():
    #         e = copy(e)
    #         e.source = translation[u]
    #         e.destination = translation[v]
    #         e.write()
    #     return cp
    
    # def __deepcopy__(self, memo : Dict[int, Any]) -> "graph":
    #     """
    #     Implements deepcopy of self
    #     """
    #     from copy import deepcopy
    #     translation = {u : deepcopy(u, memo) for u in self.vertices}
    #     cp = graph()
    #     cp.vertices.update(translation.values())
    #     for u, e, v in self.pairs():
    #         e = deepcopy(e, memo)
    #         e.source = translation[u]
    #         e.destination = translation[v]
    #         e.write()
    #     return cp

    def export(self, file : str, *, subgraph_supported : bool = False) -> None:
        """
        Writes this graph under the GEXF format into file.
        """
        # List of possible attributes to include in visuals :
        # Node Size : proportional to radius! not surface area!
        # Color : R, G, B
        # Node Shape : Any of "disc", "square", "triangle" or "diamond"
        # Edge Thickness : same as weight?
        # Edge Shape : Any of "solid", "dotted", "dashed" or "double"
        from datetime import date
        import json

        forbidden_characters : dict[int, str] = {i : "�" for i in range(32) if chr(i) not in {"\n", "\t", "\r"}}

        if not isinstance(subgraph_supported, bool):
            raise TypeError("Expected bool for subgraph_supported, got " + repr(subgraph_supported.__class__.__name__))
        try:
            with open(file, "wb") as f:
                import xml.etree.ElementTree as ET
                head = b'<?xml version="1.0" encoding="UTF-8"?>'
                root = ET.Element("gexf", attrib={"xmlns":"http://www.gexf.net/1.3", "version":"1.3", "xmlns:viz":"http://www.gexf.net/1.3/viz", "xmlns:xsi":"http://www.w3.org/2001/XMLSchema-instance", "xsi:schemaLocation":"http://www.gexf.net/1.3 http://www.gexf.net/1.3/gexf.xsd"})
                meta = ET.SubElement(root, "meta", lastmodifieddate=date.today().isoformat())
                creator = ET.SubElement(meta, "creator")
                creator.text = "Graph Builder"
                description = ET.SubElement(meta, "description")
                description.text = ""
                graph = ET.SubElement(root, "graph", mode="static")
                node_attr = ET.SubElement(graph, "attributes", **{"class" : "node", "mode" : "static"})
                node_attributes = {}
                edge_attr = ET.SubElement(graph, "attributes", **{"class" : "edge", "mode" : "static"})
                edge_attributes = {}
                nodes = ET.SubElement(graph, "nodes")
                edges = ET.SubElement(graph, "edges")
                n_ids = {u : i for i, u in enumerate(self.vertices)}
                additional_links : set[tuple[Vertex, Vertex]] = set()
                for u, i in n_ids.items():
                    d = {name : str(getattr(u, name)).translate(forbidden_characters) for name in u.__pickle_slots__ if not name.startswith("_")}
                    d["Type"] = type(u).__name__
                    if "__weakref__" in d:
                        d = {}
                    if u.parent:
                        if subgraph_supported:
                            node_i = ET.SubElement(nodes, "node", id=str(i), label=u.label.translate(forbidden_characters), pid=str(n_ids[u.parent]))
                        else:
                            node_i = ET.SubElement(nodes, "node", id=str(i), label=u.label.translate(forbidden_characters))
                            additional_links.add((u.parent, u))
                    else:
                        node_i = ET.SubElement(nodes, "node", id=str(i), label=u.label.translate(forbidden_characters))
                    attr_i = ET.SubElement(node_i, "attvalues")
                    for name, value in d.items():
                        if name not in node_attributes:
                            node_attributes[name] = ET.SubElement(node_attr, "attribute", id=name, title=name, type="string")
                        if isinstance(value, (dict, list)):
                            try:
                                svalue = json.dumps(value, indent = "\t")
                            except:
                                svalue = str(value)
                        else:
                            svalue = str(value)
                        ET.SubElement(attr_i, "attvalue", **{"for" : name, "value" : svalue.translate(forbidden_characters)})
                    size = ET.SubElement(node_i, "viz:size", value=str(u.size))
                    color = ET.SubElement(node_i, "viz:color", r=str(int(u.color.R * 255)), g=str(int(u.color.G * 255)), b=str(int(u.color.B * 255)))
                for u, e, v in self.pairs():
                    d = {name : str(getattr(e, name)).translate(forbidden_characters) for name in e.__pickle_slots__ if not name.startswith("_")}
                    d["Type"] = type(e).__name__
                    if "__weakref__" in d:
                        d = {}
                    try:
                        edge_i = ET.SubElement(edges, "edge", source=str(n_ids[u]), target=str(n_ids[v]), type=("directed" if isinstance(e, Arrow) else "undirected"), label=e.label.translate(forbidden_characters), attrib=d, weight=str(e.weight))
                    except KeyError:
                        raise
                    color = ET.SubElement(edge_i, "viz:color", r=str(int(e.color.R * 255)), g=str(int(e.color.G * 255)), b=str(int(e.color.B * 255)))
                    attr_i = ET.SubElement(edge_i, "attvalues")
                    for name, value in d.items():
                        if name not in edge_attributes:
                            edge_attributes[name] = ET.SubElement(edge_attr, "attribute", id=name, title=name, type="string")
                        if isinstance(value, (dict, list)):
                            try:
                                svalue = json.dumps(value, indent = "\t")
                            except:
                                svalue = str(value)
                        else:
                            svalue = str(value)
                        ET.SubElement(attr_i, "attvalue", **{"for" : name, "value" : svalue.translate(forbidden_characters)})
                for u, v in additional_links:
                    edge_i = ET.SubElement(edges, "edge", source=str(n_ids[u]), target=str(n_ids[v]), type="directed", label="contains", weight="1.0")
                f.write(head + b"\n")
                ET.indent(root, "\t")
                for line in ET.tostringlist(root):
                    f.write(line + b"\n")
        except Exception as E:
            raise
    

del RLock, Any, Callable, Iterable, Iterator, Optional, TypeVar, Union, Color, InstanceReferencingClass, CLS, ColorSetting, SizeSetting, SwitchSetting, WeightSetting


"""
This module contains integration protocols for this behavioral package.
"""

from ....logger import logger
from ...build import BuildingPhase
from ...utils import chrono
from . import entities, relations
from .utils import CallHandler

__all__ = []





logger.info("Loading integrations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

@chrono
def integrate_thread_creation(c : entities.Call):
    """
    Links a Thread vertex to the Call vertex (from another thread) that created it.
    """
    from ....logger import logger
    from ...graph import find_or_create
    from .entities import Thread
    from .relations import InjectedThread, StartedThread
    if c.arguments.thread_identifier != 0:      # Should indicate that the call failed
        logger.debug("New thread detected.")
        t : Thread
        t, ok = find_or_create(Thread, TID = c.arguments.thread_identifier)
        l = StartedThread(c, t)
        if c.name in {"CreateRemoteThread", "CreateRemoteThreadEx"}:
            logger.info("Detected thread injection.")
            l.remote = True
            p = c.thread.process
            InjectedThread(p, t)

@chrono
def integrate_process_creation(c : entities.Call):
    """
    Links a Process vertex to the Call vertex (from another process) that created it.
    """
    from ....logger import logger
    from ...graph import find_or_create
    from .entities import Process
    from .relations import StartedProcess
    if c.arguments.process_identifier != 0:
        logger.debug("New process detected.")
        p : Process
        p, ok = find_or_create(Process, PID = c.arguments.process_identifier)
        l = StartedProcess(c, p)
        




# Thread creation handlers
CallHandler(integrate_thread_creation, "CreateThread")          # Basic thread creation
CallHandler(integrate_thread_creation, "CreateRemoteThread", "CreateRemoteThreadEx")    # Remote thread creation (into another process)

# Process creation handlers
CallHandler(integrate_process_creation, lambda name : "CreateProcess" in name)





__N_phase = BuildingPhase.request_finalizing_phase()

@chrono
def remove_cuckoo_phantom_processes(ev : BuildingPhase):
    """
    When called with the right finalizing phase event, will cause Process nodes which represent Cuckoo process (inital processes with no info available) to be erased.
    """
    from ....logger import logger
    from ...graph import Graph
    from ..network import Host, SpawnedProcess
    from .entities import Process
    if ev.major == "Finalizer" and ev.minor == __N_phase:
        logger.debug("Erasing Cuckoo phantom processes.")
        n = 0
        h = Host.current
        for e in h.edges.copy():
            p = e.destination
            if isinstance(p, Process) and not p.parent_process and not p.executable:
                for pi in p.children_processes:
                    SpawnedProcess(h, pi)
                for ei in p.edges.copy():
                    if isinstance(e, SpawnedProcess):
                        h = e.source
                    ei.delete()
                    for g in Graph.active_graphs():
                        g.remove(ei)
                for g in Graph.active_graphs():
                    g.remove(p)
                n += 1
        if n:
            logger.debug("Actually removed {} Cuckoo process{}.".format(n, "es" if n > 1 else ""))





BuildingPhase.add_callback(remove_cuckoo_phantom_processes)





del BuildingPhase, CallHandler, chrono, entities, integrate_process_creation, integrate_thread_creation, logger, relations, remove_cuckoo_phantom_processes


"""
This module contains Vertex subclasses for this behavioral package.
"""

from collections import defaultdict
from typing import Optional

from Viper.meta.decorators import staticproperty

from ....logger import logger
from ...config import ColorSetting, SizeSetting
from ...colors import Color
from ...graph import UniqueVertex, Vertex

__all__ = ["Connection", "Socket", "Host"]





logger.info("Loading entities from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class Connection(UniqueVertex):

    """
    A connection vertex. This is a link vertex between two socket. It might only exist for a part of the lifetime of both its connected sockets.
    """

    __slots__ = {
        "duration" : "The duration that the connection was maintained for",
        "volume" : "The amount of data transfered through the connection",
        "src" : "The source address",
        "dest" : "The destination address"
    }

    __pickle_slots__ = {
        "duration",
        "volume",
        "src",
        "dest"
    }

    default_color = ColorSetting(Color(100, 50, 255))
    default_size = SizeSetting(1.5)

    def __init__(self, *, parent: Optional[Vertex] = None) -> None:
        from typing import Any
        super().__init__(parent=parent)
        self.duration : float = 0.
        self.volume : int = 0
        self.src : Any = None
        self.dest : Any = None

    @property
    def socket(self) -> "Socket":
        """
        The local Socket that this connection was made through.
        """
        from .relations import HasConnection
        for e in self.edges:
            if isinstance(e, HasConnection):
                return e.source
        raise RuntimeError("Got a Connection bound to no Socket.")





class Socket(UniqueVertex):

    """
    A socket vertex. Represents a handle given by the system to allow some kind of communication.
    """

    __slots__ = {
        "family" : "The socket address family. For example, 'InterNetwork' is for IP V4 addresses. Refer to Socket.families for documentation",
        "protocol" : "The transport protocol used by the socket. For example, Tcp. Refer to Socket.protocols for documentation",
        "type" : "The type of socket. For exemple, Dgram is a socket that supports datagrams. Refer to Socket.types for documentation"
    }

    __pickle_slots__ = {
        "family",
        "protocol",
        "type"
    }

    default_color = ColorSetting(Color(178, 153, 153))
    default_size = SizeSetting(2.5)

    families = defaultdict(lambda : ("Uncharted", "The value does not correspond to any known address family"), {
        16 : ('AppleTalk', 'AppleTalk address.'),
        22 : ('Atm', 'Native ATM services address.'),
        21 : ('Banyan', 'Banyan address.'),
        10 : ('Ccitt', 'Addresses for CCITT protocols, such as X.25.'),
        5 : ('Chaos', 'Address for MIT CHAOS protocols.'),
        24 : ('Cluster', 'Address for Microsoft cluster products.'),
        65537 : ('ControllerAreaNetwork', 'Controller Area Network address.'),
        9 : ('DataKit', 'Address for Datakit protocols.'),
        13 : ('DataLink', 'Direct data-link interface address.'),
        12 : ('DecNet', 'DECnet address.'),
        8 : ('Ecma', 'European Computer Manufacturers Association (ECMA) address.'),
        19 : ('FireFox', 'FireFox address.'),
        15 : ('HyperChannel', 'NSC Hyperchannel address.'),
        1284425 : ('Ieee', 'IEEE 1284.4 workgroup address.'),
        3 : ('ImpLink', 'ARPANET IMP address.'),
        2 : ('InterNetwork', 'Address for IP version 4.'),
        623 : ('InterNetworkV', 'Address for IP version 6.'),
        6 : ('Ipx', 'IPX or SPX address.'),
        26 : ('Irda', 'IrDA address.'),
        7 : ('Iso', 'Address for ISO protocols.'),
        14 : ('Lat', 'LAT address.'),
        29 : ('Max', 'MAX address.'),
        17 : ('NetBios', 'NetBios address.'),
        28 : ('NetworkDesigners', 'Address for Network Designers OSI gateway-enabled protocols.'),
        6 : ('NS', 'Address for Xerox NS protocols.'),
        7 : ('Osi', 'Address for OSI protocols.'),
        65536 : ('Packet', 'Low-level Packet address.'),
        4 : ('Pup', 'Address for PUP protocols.'),
        11 : ('Sna', 'IBM SNA address.'),
        1 : ('Unix', 'Unix local to host address.'),
        -1 : ('Unknown', 'Unknown address family.'),
        0 : ('Unspecified', 'Unspecified address family.'),
        18 : ('VoiceView', 'VoiceView address.'),
    })

    protocols = defaultdict(lambda : ("Uncharted", "The value does not correspond to any known protocol"), {
        3 : ('Ggp', 'Gateway To Gateway Protocol.'),
        1 : ('Icmp', 'Internet Control Message Protocol.'),
        58 : ('IcmpV6', 'Internet Control Message Protocol for IPv6.'),
        22 : ('Idp', 'Internet Datagram Protocol.'),
        2 : ('Igmp', 'Internet Group Management Protocol.'),
        0 : ('IP', 'Internet Protocol.'),
        51 : ('IPSecAuthenticationHeader', 'IPv6 Authentication header. For details'),
        50 : ('IPSecEncapsulatingSecurityPayload', 'IPv6 Encapsulating Security Payload header.'),
        4 : ('IPv4', 'Internet Protocol version 4.'),
        41 : ('IPv6', 'Internet Protocol version 6 (IPv6).'),
        60 : ('IPv6DestinationOptions', 'IPv6 Destination Options header.'),
        44 : ('IPv6FragmentHeader', 'IPv6 Fragment header.'),
        0 : ('IPv6HopByHopOptions', 'IPv6 Hop by Hop Options header.'),
        59 : ('IPv6NoNextHeader', 'IPv6 No next header.'),
        43 : ('IPv6RoutingHeader', 'IPv6 Routing header.'),
        1000 : ('Ipx', 'Internet Packet Exchange Protocol.'),
        77 : ('ND', 'Net Disk Protocol (unofficial).'),
        12 : ('Pup', 'PARC Universal Packet Protocol.'),
        255 : ('Raw', 'Raw IP packet protocol.'),
        1256 : ('Spx', 'Sequenced Packet Exchange protocol.'),
        1257 : ('SpxII', 'Sequenced Packet Exchange version 2 protocol.'),
        6 : ('Tcp', 'Transmission Control Protocol.'),
        17 : ('Udp', 'User Datagram Protocol.'),
        -1 : ('Unknown', 'Unknown protocol.'),
        0 : ('Unspecified', 'Unspecified protocol.'),
    })

    types = defaultdict(lambda : ("Uncharted", "The value does not correspond to any known socket type"), {
        2 : ('Dgram', 'Supports datagrams, which are connectionless, unreliable messages of a fixed (typically small) maximum length. Messages might be lost or duplicated and might arrive out of order. A Socket of type Dgram requires no connection prior to sending and receiving data, and can communicate with multiple peers. Dgram uses the Datagram Protocol (ProtocolType.Udp) and the AddressFamily. InterNetwork address family.'),
        3 : ('Raw', 'Supports access to the underlying transport protocol. Using Raw, you can communicate using protocols like Internet Control Message Protocol (ProtocolType.Icmp) and Internet Group Management Protocol (ProtocolType.Igmp). Your application must provide a complete IP header when sending. Received datagrams return with the IP header and options intact.'),
        4 : ('Rdm', 'Supports connectionless, message-oriented, reliably delivered messages, and preserves message boundaries in data. Rdm (Reliably Delivered Messages) messages arrive unduplicated and in order. Furthermore, the sender is notified if messages are lost. If you initialize a Socket using Rdm, you do not require a remote host connection before sending and receiving data. With Rdm, you can communicate with multiple peers.'),
        5 : ('Seqpacket', 'Provides connection-oriented and reliable two-way transfer of ordered byte streams across a network. Seqpacket does not duplicate data, and it preserves boundaries within the data stream. A Socket of type Seqpacket communicates with a single peer and requires a remote host connection before communication can begin.'),
        1 : ('Stream', 'Supports reliable, two-way, connection-based byte streams without the duplication of data and without preservation of boundaries. A Socket of this type communicates with a single peer and requires a remote host connection before communication can begin. Stream uses the Transmission Control Protocol (ProtocolType.Tcp) and the AddressFamily.InterNetwork address family.'),
        -1 : ('Unknown', 'Specifies an unknown Socket type.'),
    })

    def __init__(self, *, parent: Optional[Vertex] = None) -> None:
        super().__init__(parent=parent)
        self.family : str = ""
        self.protocol : str = ""
        self.type : str = ""





class Host(UniqueVertex):

    """
    A machine vertex. It represents a physical machine.
    """

    __current : Optional["Host"] = None

    @staticproperty
    def current() -> "Host":
        """
        The currently active Host node. This is used to indicate which Host node is the machine running the sample.
        """
        if Host.__current is None:
            raise AttributeError("Host class has no attribute 'current'.")
        return Host.__current

    @current.setter
    def current(value : "Host"):
        if not isinstance(value, Host):
            raise TypeError("Cannot set attribute 'current' of class 'Host' to object of type '{}'".format(type(value).__name__))
        Host.__current = value

    __slots__ = {
        "address" : "The IP address of the machine",
        "domain" : "The URL the machine is known as if any",
        "name" : "The machine's name",
        "platform" : "The operating system the host is running on"
    }

    __pickle_slots__ = {
        "address",
        "domain",
        "name",
        "platform"
    }

    default_color = ColorSetting(Color.white)
    default_size = SizeSetting(10.0)

    def __init__(self, *, parent: Vertex | None = None) -> None:
        super().__init__(parent=parent)
        self.address : str = ""
        self.domain : str = ""
        self.name : str = ""
        self.platform : str = "Unknown"
    
    @property
    def label(self) -> str:
        """
        Returns a label for this Host node.
        """
        if self.name:
            return "Host " + repr(self.name)
        return "Host at " + self.address





del Color, ColorSetting, SizeSetting, Optional, UniqueVertex, Vertex, defaultdict, logger, staticproperty


"""
This module contains integration protocols for this behavioral package.
"""

from ....logger import logger
from ...utils import chrono
from ..execution.entities import Call
from ..execution.utils import CallHandler
from . import entities, relations

__all__ = []





logger.info("Loading integrations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

__active_connections : dict[entities.Socket, entities.Connection] = {}
__active_handles : dict[entities.Socket, int] = {}
__inverted_handles : dict[int, entities.Socket] = {}

@chrono
def integrate_socket_creation(c : Call):
    """
    Creates a Socket vertex and links it to the call that created it.
    """
    from ....logger import logger
    from .entities import Socket
    from .relations import CreatesSocket, HasSocket
    if c.status == 1:
        logger.debug("Socket created.")
        s = Socket()
        __active_handles[s] = c.arguments.socket
        __inverted_handles[c.arguments.socket] = s
        s.family = Socket.families[c.arguments.af][0]
        s.protocol = Socket.protocols[c.arguments.protocol][0]
        s.type = Socket.types[c.arguments.type][0]
        if s in __active_connections:
            __active_connections.pop(s)
        CreatesSocket(c, s)
        HasSocket(c.thread.process, s)

@chrono
def integrate_socket_binding(c : Call):
    """
    Creates a Connection vertex and binds it to a local address.
    """
    from ....logger import logger
    from ...graph import find_or_create
    from .entities import Connection, Host
    from .relations import Binds, Communicates, HasConnection
    if c.status == 1:
        logger.debug("Socket bound.")
        l = Connection()
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Binding to non-existing socket.")
            return
        __active_connections[s] = l
        HasConnection(s, l)
        Communicates(l, find_or_create(Host, domain = "host")[0])
        b = Binds(c, l)
        if s.family == "InterNetwork":
            b.src = (c.arguments.ip_address, c.arguments.port)
            l.src = b.src
        else:
            logger.error("I don't know how to handle this type of address:\n{}".format(s))

@chrono
def integrate_socket_connection(c : Call):
    """
    Sets the remote address of a connection.
    """
    from ....logger import logger
    from ...graph import find_or_create
    from .entities import Connection, Host
    from .relations import Communicates, Connects, HasConnection
    if c.status == 1:
        logger.debug("Socket connected.")
        try:
            sock = c.arguments.socket
        except AttributeError:
            sock = c.arguments.s
        if sock in __inverted_handles:
            s = __inverted_handles[sock]
        else:
            logger.warning("Connecting from non-existing socket.")
            return
        if s in __active_connections:
            l = __active_connections[s]
            logger.warning("Trying to connect an unbound socket...")
        else:
            l = Connection()
        HasConnection(s, l)
        Communicates(l, find_or_create(Host, domain = "host")[0])
        __active_connections[s] = l
        b = Connects(c, l)
        Communicates(l, find_or_create(Host, address = c.arguments.ip_address)[0])
        if s.family == "InterNetwork":
            b.dest = (c.arguments.ip_address, c.arguments.port)
            l.dest = b.dest
        else:
            logger.error("I don't know how to handle this type of address:\n{}".format(s))


@chrono
def integrate_socket_listening(c : Call):
    """
    Sets the socket in listening mode.
    """
    from ....logger import logger
    from .relations import ListensOn
    if c.status == 1:
        logger.debug("Socket listening.")
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Listening from non-existing socket.")
            return
        ListensOn(c, s)

@chrono
def integrate_socket_accepting(c : Call):
    """
    Sets a connection remote address through an accept system call. Also creates a new socket object.
    """
    from ....logger import logger
    from ...graph import find_or_create
    from .entities import Connection, Host, Socket
    from .relations import Accepts, Communicates, Duplicates, HasConnection
    if c.status == 1:
        logger.debug("Socket accepting connection.")
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Accepting from non-existing socket.")
            return
        s1 = Socket()
        l1 = Connection()
        s1.family = s.family
        s1.protocol = s.protocol
        s1.type = s.type
        __active_handles[s1] = c.return_value
        __inverted_handles[c.return_value] = s1
        if s not in __active_connections:
            logger.warning("Accepting from unbound socket...")
        else:
            l = __active_connections[s]
            l1.src = l.src
        b = Accepts(c, l1)
        HasConnection(s1, l1)
        Duplicates(s, s1)
        Communicates(l1, find_or_create(Host, address = c.arguments.ip_address)[0])
        if s1.family == "InterNetwork":
            b.dest = (c.arguments.ip_address, c.arguments.port)
            l1.dest = b.dest

@chrono
def integrate_socket_close(c : Call):
    """
    Closes a socket and the connection linked to it.
    """
    from ....logger import logger
    from .relations import Closes, CloseSocket
    if c.status == 1:
        logger.debug("Closing socket.")
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Closing non-existing socket.")
            return
        CloseSocket(c, s)
        handle = __active_handles.pop(s)
        __inverted_handles.pop(handle)
        if s in __active_connections:
            l = __active_connections.pop(s)
            Closes(c, l)
        else:
            logger.warning("Closing down unbound socket...")

@chrono
def integrate_socket_shutdown(c : Call):
    """
    Closes a connection.
    """
    from ....logger import logger
    from .relations import Shutdown
    if c.status == 1:
        logger.debug("Closing socket.")
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Shuting down non-existing socket")
            return
        if s in __active_connections:
            l = __active_connections.pop(s)
            Shutdown(c, l)
        else:
            logger.warning("Shuting down unbound socket...")

@chrono
def integrate_socket_send(c : Call):
    """
    Sends data through a connection.
    """
    from ....logger import logger
    from ..data import Data
    from ..data.integration import register_write_operation
    from .relations import Conveys, Sends
    if c.status == 1:
        logger.debug("Sending data.")
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Sending through non-existing socket.")
            return
        if s not in __active_connections:
            logger.warning("Sending through unbound socket.")
        else:
            l = __active_connections[s]
            d = Data()
            d.time = c.time
            d.set_data(c.arguments.buffer)
            Sends(c, d)
            Conveys(d, l)
            l.volume += len(d.data)
            register_write_operation(l, s, c.arguments.buffer)

@chrono
def integrate_socket_recv(c : Call):
    """
    Receives data through a connection.
    """
    from ....logger import logger
    from ..data import Data
    from ..data.integration import register_read_operation
    from .relations import Conveys, Receives
    if c.status == 1:
        logger.debug("Receiving data.")
        if c.arguments.socket in __inverted_handles:
            s = __inverted_handles[c.arguments.socket]
        else:
            logger.warning("Receiving through non-existing socket.")
            return
        if s not in __active_connections:
            logger.warning("Receiving through unbound socket.")
        else:
            l = __active_connections[s]
            d = Data()
            d.time = c.time
            d.set_data(c.arguments.buffer)
            Receives(d, c)
            Conveys(l, d)
            l.volume += len(d.data)
            register_read_operation(l, s, c.arguments.buffer)





# Socket creation
CallHandler(integrate_socket_creation, "socket", "WSASocketW")

# Socket binding
CallHandler(integrate_socket_binding, "bind")

# Socket connecting
CallHandler(integrate_socket_connection, "connect", "WSAConnect")

# Socket listening
CallHandler(integrate_socket_listening, "listen")

# Socket accepting
CallHandler(integrate_socket_accepting, "accept")

# Socket closing
CallHandler(integrate_socket_close, "closesocket")

# Socket shuting down
CallHandler(integrate_socket_shutdown, "shutdown")

# Socket sending
CallHandler(integrate_socket_send, "send")

# Socket receiving
CallHandler(integrate_socket_recv, "recv")





del Call, CallHandler, chrono, entities, integrate_socket_accepting, integrate_socket_binding, integrate_socket_close, integrate_socket_connection, integrate_socket_creation, integrate_socket_listening, integrate_socket_recv, integrate_socket_send, integrate_socket_shutdown, logger, relations


"""
This module contains Vertex subclasses for this behavioral package.
"""

from ....logger import logger
from ...config import ColorSetting, SizeSetting
from ...colors import Color
from ...graph import UniqueVertex, Vertex

__all__ = ["Import"]





logger.info("Loading entities from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class Import(UniqueVertex):
    
    """
    An import vertex. Shows the use of a DLL or equivalent.
    """

    __slots__ = {
        "__name" : "The name of the import file.",
        "path" : "The path to the file.",
        "length" : "The size in bytes of the file."
    }

    __pickle_slots__ = {
        "name",
        "path",
        "length"
    }

    __suspicious_keyword_names = {
        "crypt",
        "advapi",
        "kernel",
        "sock"
    }

    default_color = ColorSetting(Color(150, 150, 0))
    default_size = SizeSetting(0.75)

    suspicious_import_color = ColorSetting(Color(255, 150, 0))

    def __init__(self, *, parent: Vertex | None = None) -> None:
        super().__init__(parent=parent)
        self.__name : str = ""
        self.path : str = ""
        self.length : int = 0

    @property
    def name(self) -> str:
        """
        The name of the imported library.
        """
        return self.__name

    @name.setter
    def name(self, n : str):
        from ...config import ColorSetting
        if not isinstance(n, str):
            raise TypeError("Expected str, got " + repr(type(n).__name__))
        self.__name = n
        self.color = self.default_color
        for kw in self.__suspicious_keyword_names:
            if kw in n.lower():
                self.color = self.suspicious_import_color
                break
    
    @property
    def label(self) -> str:
        """
        Returns a label for this Import node.
        """
        return "Import {}".format(self.__name.lower())
    




del Color, ColorSetting, SizeSetting, UniqueVertex, Vertex, logger


"""
This module contains Edge and Arrow subclasses for this behavioral package.
"""

from ....logger import logger
from ...graph import Arrow, Edge
from ..data.entities import Data
from ..execution.entities import Call, Process
from ..network.entities import Host
from .entities import Directory, File, Handle

__all__ = ["UsesFile", "UsesDirectory", "Opens", "Closes", "CreatesFile", "CreatesDirectory", "DeletesFile", "DeletesDirectory", "Reads", "Writes", "Conveys", "HasDrive", "HasHandle", "Contains", "IsCopyiedInto"]





logger.info("Loading relations from {} library.".format(__name__.rpartition(".")[0].rpartition(".")[2]))

class UsesFile(Edge):

    """
    This kind of edge indicates that a handle uses a file.
    """

    label : str = ""

    source : Handle
    destination : File





class UsesDirectory(Edge):

    """
    This kind of edge indicates that a handle uses a directory.
    """

    label : str = ""

    source : Handle
    destination : Directory





class Opens(Edge):

    """
    This kind of edge indicates that a system call opened a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class Closes(Edge):

    """
    This kind of edge indicates that a system call closed a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class CreatesFile(Edge):

    """
    This kind of edge indicates that a system call created a file through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class CreatesDirectory(Edge):

    """
    This kind of edge indicates that a system call created a directory through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class DeletesFile(Edge):

    """
    This kind of edge indicates that a system call deleted a file through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class DeletesDirectory(Edge):

    """
    This kind of edge indicates that a system call deleted a directory through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class Reads(Arrow):

    """
    This kind of arrow indicates that a system call read data from a file.
    """

    source : Data
    destination : Call





class Writes(Arrow):

    """
    This kind of arror indicates that a system call wrote data to a file.
    """

    source : Call
    destination : Data





class Conveys(Arrow):

    """
    This kind of arrow indicates a data flux with a file handle.
    """

    label : str = ""

    source : Handle | Data
    destination : Handle | Data





class HasDrive(Edge):

    """
    This kind of edge indicates that a machine owns a disk drive/partition(/root directory).
    """

    label : str = ""

    source : Host
    destination : Directory





class HasHandle(Edge):

    """
    This kind of edge indicates that a process owns a file handle.
    """

    label : str = ""

    source : Process
    destination : Handle





class Contains(Arrow):

    """
    This kind of arrow indicates that a directory contains a file or another directory.
    """

    label : str = ""

    source : Directory
    destination : Directory | File





class IsCopyiedInto(Arrow):

    """
    This kind indicates that the destination file is a copy of the source file.
    """

    source : File
    destination : File





del Arrow, Call, Data, Directory, Edge, File, Handle, Host, Process, logger